from typing import Type, Optional, Dict, Any, List
from pydantic import BaseModel, Field
from mmh3 import hash_bytes
import base64
import json
import os
from langchain_llm_utils.common import get_logger, BaseModelType

logger = get_logger("RequestCache")


class RequestCache:
    """
    Cache for storing and retrieving structured responses.

    This class provides a persistent cache for storing LLM responses, with support for:
    1. Storing and retrieving Pydantic models
    2. Efficient key hashing for large prompts
    3. Persistent storage in JSON format
    4. Both raw and parsed data access

    Example usage:

    ```python
    # Basic usage
    cache = RequestCache("responses_cache.json")

    # With LLM
    from llm_utils.llm import LLM

    class TranslationResponse(BaseModel):
        translation: str
        source_language: str

    llm = LLM(
        model="gemini-1.5-flash",
        cache=cache,
        response_model=TranslationResponse
    )

    # Direct cache access
    response = cache.get_parsed("some_key", TranslationResponse)
    if response:
        print(f"Cached translation: {response.translation}")

    # Storing new response
    new_response = TranslationResponse(
        translation="Hello",
        source_language="English"
    )
    cache.set("some_key", new_response)

    # Save changes to disk
    cache.save_cache()

    # Force overwrite existing cache file
    cache.save_cache(overwrite=True)
    ```

    """

    def __init__(self, cache_file: str):
        """
        Initialize the cache.

        Args:
            cache_file: Path to the JSON file where cache will be stored
                If file doesn't exist, an empty cache will be created
                If file exists, it will be loaded

        Example:
            ```python
            # Create/load cache
            cache = RequestCache("my_cache.json")

            # Create cache in specific directory
            cache = RequestCache("data/caches/responses.json")
            ```
        """
        self.cache_file = cache_file
        self.cache = self._load_cache()

    def _load_cache(self) -> Dict:
        try:
            with open(self.cache_file, "r") as f:
                logger.info(f"Loading cache from {self.cache_file}")
                return json.load(f)
        except FileNotFoundError:
            logger.info(f"Cache file {self.cache_file} not found, creating one ..")
            # Try to create the file and return an empty dict
            try:
                with open(self.cache_file, "w") as f:
                    return {}
            except Exception as e:
                logger.error(f"Failed to create cache file {self.cache_file}: {e}")
                return {}

    def _hash_key(self, key: str) -> str:
        """
        Generate a deterministic hash for a cache key.

        Args:
            key: The original key string

        Returns:
            A base64-encoded MurmurHash3 hash of the key
        """
        # Get bytes hash from MurmurHash3
        hash_value = hash_bytes(key.encode("utf-8"))
        # Convert to base64 for readability and storage efficiency
        return base64.urlsafe_b64encode(hash_value).decode("ascii")

    def get(self, key: str) -> Optional[Dict]:
        """
        Get raw cached value.

        Args:
            key: Cache key to lookup

        Returns:
            Raw cached dictionary or None if key not found
        """
        hashed_key = self._hash_key(key)
        # logger.debug(f"Hashed key: {hashed_key}")
        # logger.debug(f"Result: {self.cache.get(hashed_key)}")
        return self.cache.get(hashed_key)

    def get_cache_key(self, **kwargs) -> str:
        """
        Get the cache key for a given key.

        Args:
            kwargs: Keyword arguments to include in the cache key

        Returns:
            A base64-encoded MurmurHash3 hash of the key
        """
        # Create : delimited string from kwargs
        key_str = ":".join([f"{k}__{v}" for k, v in kwargs.items()])
        return self._hash_key(key_str)

    def get_parsed(
        self, key: str, model: Type[BaseModelType]
    ) -> Optional[BaseModelType]:
        """
        Get a cached value and parse it into a Pydantic model.

        Args:
            key: Cache key to lookup
            model: Pydantic model class to parse the cached data into

        Returns:
            Parsed Pydantic model instance or None if key not found
        """
        cached_result = self.get(key)
        if cached_result is None:
            return None

        def recursive_construct(cls, data):
            if isinstance(data, dict):
                for field_name, field_value in data.items():
                    field_info = cls.model_fields.get(field_name)
                    if (
                        field_info
                        and isinstance(field_info.annotation, type)
                        and issubclass(field_info.annotation, BaseModel)
                    ):
                        data[field_name] = recursive_construct(
                            field_info.annotation, field_value
                        )
                    elif (
                        field_info
                        and isinstance(field_info.annotation, type)
                        and issubclass(field_info.annotation, List)
                    ):
                        if (
                            hasattr(field_info.annotation, "__args__")
                            and isinstance(field_info.annotation.__args__[0], type)
                            and issubclass(field_info.annotation.__args__[0], BaseModel)
                        ):
                            data[field_name] = [
                                recursive_construct(
                                    field_info.annotation.__args__[0], item
                                )
                                for item in field_value
                            ]
            return cls(**data)

        try:
            return model.model_validate(cached_result)
        except Exception as e:
            logger.error(
                f"Failed to parse cached result with pydantic v2 method, trying recursive construct: {e}"
            )
            return recursive_construct(model, cached_result)

    def set(self, key: str, value: BaseModel) -> None:
        """
        Store a Pydantic model in the cache.

        Args:
            key: Cache key
            value: Pydantic model instance to store
        """
        hashed_key = self._hash_key(key)
        if isinstance(value, BaseModel):
            self.cache[hashed_key] = value.model_dump()
        else:
            self.cache[hashed_key] = value

    def clear(self) -> None:
        """Clear the cache."""
        self.cache = {}
        self.save_cache(overwrite=True)

    def save_cache(self, overwrite: bool = False) -> None:
        """Save the cache to disk."""
        logger.info(f"Saving cache to {self.cache_file}")
        file_exists = os.path.exists(self.cache_file)

        if overwrite or not file_exists:
            with open(self.cache_file, "w") as f:
                json.dump(self.cache, f)



import uuid
from langchain_llm_utils.common import decider


def test_decider_edge_cases():
    test_uuid = uuid.uuid4()

    # Should never sample when rate is 0
    assert not decider(test_uuid, 0.0)
    assert not decider(test_uuid, -0.1)

    # Should always sample when rate is 1 or greater
    assert decider(test_uuid, 1.0)
    assert decider(test_uuid, 1.1)


def test_decider_deterministic():
    # Same inputs should give same results
    test_uuid = uuid.uuid4()
    result1 = decider(test_uuid, 0.5)
    result2 = decider(test_uuid, 0.5)
    assert result1 == result2


def test_decider_distribution():
    # Generate many UUIDs and verify sampling rate is approximately correct
    sample_rate = 0.3
    num_trials = 10000
    sampled = 0

    for _ in range(num_trials):
        if decider(uuid.uuid4(), sample_rate):
            sampled += 1

    # Calculate actual rate
    actual_rate = sampled / num_trials

    # Allow for some statistical variance (within ±2 percentage points)
    assert (
        abs(actual_rate - sample_rate) < 0.02
    ), f"Expected rate ~{sample_rate}, got {actual_rate}"



import pytest
from unittest.mock import patch, MagicMock, AsyncMock
from pydantic import BaseModel, Field
from typing import List, Callable
from langchain_llm_utils.llm import LLM, ModelProvider


class JokeResponse(BaseModel):
    setup: str = Field(default="What do you call a fish with no eyes?")
    punchline: str = Field(default="A fsh!")


@pytest.fixture
def mock_llm_provider():
    """Fixture to mock the LLM provider."""
    with patch(
        "langchain_llm_utils.llm.LLMFactory.create_provider"
    ) as mock_create_provider:
        mock_provider = MagicMock()
        # Set up sync mock
        mock_provider.invoke.return_value = "Mocked response"

        # Set up async mock
        async_mock = AsyncMock()
        async_mock.return_value = "Mocked async response"
        mock_provider.ainvoke = async_mock

        mock_create_provider.return_value = mock_provider
        yield mock_provider


@pytest.fixture
def llm_instance(mock_llm_provider):
    """Fixture to create an LLM instance for testing."""
    return LLM(model_provider=ModelProvider.OPENAI, model_name="gpt-4o")


@pytest.fixture
def llm_instance_with_response_model(mock_llm_provider):
    """Fixture to create an LLM instance with a response model for testing structured output."""
    return LLM(
        model_provider=ModelProvider.OPENAI,
        model_name="gpt-4o",
        response_model=JokeResponse,
    )


def test_invoke(mock_llm_provider, llm_instance):
    """Test the invoke method."""
    # Mock the _llm.invoke method to return a specific response
    mock_llm_provider.invoke.return_value = "Mocked response"

    response = llm_instance.invoke("What is the capital of France?")
    assert response == "Mocked response"
    mock_llm_provider.invoke.assert_called_once_with("What is the capital of France?")


def test_invoke_connection_error(mock_llm_provider, llm_instance):
    """Test that connection errors are re-raised in invoke method."""
    # Mock the provider to raise a ConnectionRefusedError
    mock_llm_provider.invoke.side_effect = ConnectionRefusedError("Connection refused")

    with pytest.raises(ConnectionRefusedError) as exc_info:
        llm_instance.invoke("Test prompt")
    assert "Failed to connect to LLM" in str(exc_info.value)


def test_invoke_other_exception(mock_llm_provider, llm_instance):
    """Test that non-connection errors return None in invoke method."""
    # Mock the provider to raise a generic exception
    mock_llm_provider.invoke.side_effect = Exception("Some other error")

    response = llm_instance.invoke("Test prompt")
    assert response is None


def test_generate(mock_llm_provider, llm_instance):
    """Test the generate method."""
    # Mock the _llm.invoke method to return a specific response
    mock_llm_provider.invoke.return_value = "Mocked response"

    response = llm_instance.generate("Tell me a joke")
    assert response == "Mocked response"
    mock_llm_provider.invoke.assert_called_once_with("Tell me a joke")


def test_generate_structured_output(
    llm_instance_with_response_model, mock_llm_provider
):
    """Test the generate method with structured output."""
    # Mock the _llm.invoke method to return a structured response
    mock_response = JokeResponse()
    mock_llm_provider.invoke.return_value = mock_response

    response = llm_instance_with_response_model.generate("Tell me a joke")

    assert isinstance(response, JokeResponse)
    assert response.setup == mock_response.setup
    assert response.punchline == mock_response.punchline


def test_generate_structured_output_when_no_response_from_llm(
    llm_instance_with_response_model, mock_llm_provider
):
    """Test the generate method with structured output."""
    # Mock the _llm.invoke method to return None (failed LLM call)
    mock_llm_provider.invoke.return_value = None

    response = llm_instance_with_response_model.generate("Tell me a joke")
    assert isinstance(
        response, JokeResponse
    )  # Make sure response is of type JokeResponse (default response model)


def test_generate_with_template(mock_llm_provider, llm_instance):
    """Test the generate method with a template and input variables."""
    # Mock the _llm.invoke method to return a specific response
    mock_llm_provider.invoke.return_value = "Mocked response"

    response = llm_instance.generate(
        template="Tell me a {type} joke about {topic}",
        input_variables={"type": "dad", "topic": "programming"},
    )
    assert (
        response == "Mocked response"
    )  # Check if the response matches the mocked response
    mock_llm_provider.invoke.assert_called_once_with(
        "Tell me a dad joke about programming"
    )


@pytest.mark.asyncio
async def test_ainvoke(mock_llm_provider, llm_instance):
    """Test the async invoke method."""
    # Mock the _llm.ainvoke method to return a specific response
    mock_llm_provider.ainvoke.return_value = "Mocked async response"

    response = await llm_instance.ainvoke("What is the capital of France?")
    assert response == "Mocked async response"
    mock_llm_provider.ainvoke.assert_called_once_with("What is the capital of France?")


@pytest.mark.asyncio
async def test_ainvoke_connection_error(mock_llm_provider, llm_instance):
    """Test that connection errors are re-raised in async invoke method."""
    # Mock the provider to raise a ConnectionRefusedError
    mock_llm_provider.ainvoke.side_effect = ConnectionRefusedError("Connection refused")

    with pytest.raises(ConnectionRefusedError) as exc_info:
        await llm_instance.ainvoke("Test prompt")
    assert "Failed to connect to LLM" in str(exc_info.value)


@pytest.mark.asyncio
async def test_ainvoke_other_exception(mock_llm_provider, llm_instance):
    """Test that non-connection errors return None in async invoke method."""
    # Mock the provider to raise a generic exception
    mock_llm_provider.ainvoke.side_effect = Exception("Some other error")

    response = await llm_instance.ainvoke("Test prompt")
    assert response is None


@pytest.mark.asyncio
async def test_agenerate(mock_llm_provider, llm_instance):
    """Test the async generate method."""
    # Mock the _llm.ainvoke method to return a specific response
    mock_llm_provider.ainvoke.return_value = "Mocked async response"

    response = await llm_instance.agenerate("Tell me a joke")
    assert response == "Mocked async response"
    mock_llm_provider.ainvoke.assert_called_once_with("Tell me a joke")


@pytest.mark.asyncio
async def test_agenerate_structured_output(
    llm_instance_with_response_model, mock_llm_provider
):
    """Test the async generate method with structured output."""
    # Mock the _llm.ainvoke method to return a structured response
    mock_response = JokeResponse()
    mock_llm_provider.ainvoke.return_value = mock_response

    response = await llm_instance_with_response_model.agenerate("Tell me a joke")

    assert isinstance(response, JokeResponse)
    assert response.setup == mock_response.setup
    assert response.punchline == mock_response.punchline


@pytest.mark.asyncio
async def test_agenerate_structured_output_when_no_response_from_llm(
    llm_instance_with_response_model, mock_llm_provider
):
    """Test the async generate method with structured output when LLM returns None."""
    # Mock the _llm.ainvoke method to return None (failed LLM call)
    mock_llm_provider.ainvoke.return_value = None

    response = await llm_instance_with_response_model.agenerate("Tell me a joke")
    assert isinstance(
        response, JokeResponse
    )  # Make sure response is of type JokeResponse (default response model)


@pytest.mark.asyncio
async def test_agenerate_with_template(mock_llm_provider, llm_instance):
    """Test the async generate method with a template and input variables."""
    # Mock the _llm.ainvoke method to return a specific response
    mock_llm_provider.ainvoke.return_value = "Mocked async response"

    response = await llm_instance.agenerate(
        template="Tell me a {type} joke about {topic}",
        input_variables={"type": "dad", "topic": "programming"},
    )
    assert response == "Mocked async response"
    mock_llm_provider.ainvoke.assert_called_once_with(
        "Tell me a dad joke about programming"
    )



from pydantic import BaseModel, Field
from langchain_llm_utils.llm import LLM, ModelProvider
from langchain_llm_utils.evaluator import BasicEvaluationSuite
from langchain_llm_utils.common import get_logger

logger = get_logger("TextAnalytics")

import argparse
import time
import asyncio


class LanguageDetected(BaseModel):
    language: str = Field(
        description="The language of the text in lowercase", default_factory=lambda: ""
    )


class Translation(BaseModel):
    translated_text: str = Field(
        description="The text after being translated to English",
        default_factory=lambda: "",
    )


class TranslationService:
    def __init__(self):
        self.evaluator = BasicEvaluationSuite(sample_rate=0.9)
        self.lang_detector = LLM(
            model_provider=ModelProvider.OPENAI,
            model_name="gpt-4o-mini",
            temperature=1.0,
            response_model=LanguageDetected,
            evaluator=self.evaluator,
            model_type="lang_detector",
            langfuse_tags=["text_analytics", "translation_service", "lang_detector"],
        )
        self.translator = LLM(
            model_provider=ModelProvider.OPENAI,
            model_name="gpt-4o-mini",
            temperature=1.0,
            response_model=Translation,
            evaluator=self.evaluator,
            model_type="translator",
            langfuse_tags=["text_analytics", "translation_service", "translator"],
        )
        self.lang_detector_prompt = """
        You are a language detector. You are given a text and you need to determine the language of the text.
        Language code can be like - en, es, fr, de, it, etc.
        Return in JSON format:
        {{
            "language": "language_code"
        }}
        
        The text is: {text}
        """
        self.translator_prompt = """
        You are a language translator. You are given a text and you need to translate it to English.
        
        Return in JSON format:
        {{
            "translated_text": "translated_text"
        }}
        
        The text is: {text}
        """


def translate_to_english(text: str) -> Translation:
    """Synchronous translation"""
    translation_service = TranslationService()

    lang_detected = translation_service.lang_detector.generate(
        template=translation_service.lang_detector_prompt,
        input_variables={"text": text},
    )
    logger.debug(f"Language detected: {lang_detected.language}")
    if lang_detected.language == "en":
        return Translation(translated_text=text)
    else:
        return translation_service.translator.generate(
            template=translation_service.translator_prompt,
            input_variables={"text": text},
        )


async def translate_to_english_async(text: str) -> Translation:
    """Asynchronous translation"""
    translation_service = TranslationService()

    # First detect language asynchronously
    lang_detected = await translation_service.lang_detector.agenerate(
        template=translation_service.lang_detector_prompt,
        input_variables={"text": text},
    )

    logger.debug(f"Language detected: {lang_detected.language}")
    if lang_detected.language == "en":
        return Translation(translated_text=text)
    else:
        # Then translate asynchronously
        return await translation_service.translator.agenerate(
            template=translation_service.translator_prompt,
            input_variables={"text": text},
        )


async def main_async():
    """
    Example command:
    poetry run python text_analytics.py "Hello, how are you?" --async
    """
    parser = argparse.ArgumentParser(description="Translate text to English")
    parser.add_argument("text", type=str, help="Text to translate")
    parser.add_argument("--async", action="store_true", help="Run asynchronously")
    args = parser.parse_args()

    if getattr(args, "async"):
        print("\n=== Running Asynchronous Translation ===")
        start_time = time.time()
        result = await translate_to_english_async(args.text)
        end_time = time.time()
        # Wait for evaluations to complete
        await asyncio.sleep(2)
        print(f"Async translated text: {result.translated_text}")
        print(f"Async time taken: {end_time - start_time:.2f} seconds")
    else:
        print("\n=== Running Synchronous Translation ===")
        start_time = time.time()
        result = translate_to_english(args.text)
        end_time = time.time()
        print(f"Sync translated text: {result.translated_text}")
        print(f"Sync time taken: {end_time - start_time:.2f} seconds")
        print("Waiting for sync evaluations to complete...")


if __name__ == "__main__":
    asyncio.run(main_async())



import pytest
import asyncio
import time
from unittest.mock import Mock, patch, MagicMock
from langchain_llm_utils.rate_limiter import (
    SmartRateLimiter,
    GPT4Tokenizer,
    LangchainTokenAwareRateLimiter,
)
from langchain_openai import ChatOpenAI
from langchain_core.messages import AIMessage
from langchain_core.outputs import ChatResult, ChatGeneration
from concurrent.futures import ThreadPoolExecutor


@pytest.fixture
def mock_tokenizer():
    """Fixture to mock the tokenizer."""
    tokenizer = Mock()
    tokenizer.return_value = 10  # Default token count for testing
    return tokenizer


@pytest.fixture
def rate_limiter(mock_tokenizer):
    """Fixture to create a rate limiter with mocked tokenizer."""
    return SmartRateLimiter(
        requests_per_minute=60,
        tokens_per_minute=4000,
        tokenizer=mock_tokenizer,
        check_every_n_seconds=0.1,
    )


@pytest.mark.asyncio
async def test_aacquire_immediate_capacity(rate_limiter):
    """Test async acquire when capacity is immediately available."""

    async with asyncio.timeout(1):  # Add timeout
        result = await rate_limiter.aacquire("test prompt")
        assert result is True


def test_acquire_sync(rate_limiter):
    """Test synchronous acquire."""
    result = rate_limiter.acquire("test prompt")
    assert result is True


def test_acquire_sync_non_blocking(rate_limiter):
    """Test synchronous acquire with non-blocking."""
    rate_limiter.request_bucket.tokens = 0
    result = rate_limiter.acquire("test prompt", blocking=False)
    assert result is False


def test_metrics(rate_limiter):
    """Test metrics collection and reporting."""
    rate_limiter.acquire("test1")
    rate_limiter.acquire("test2")
    metrics = rate_limiter.get_metrics()
    assert "available_requests" in metrics
    assert metrics["total_requests_processed"] == 2


def test_token_counting(mock_tokenizer):
    """Test token counting functionality."""
    rate_limiter = SmartRateLimiter(
        requests_per_minute=60, tokens_per_minute=4000, tokenizer=mock_tokenizer
    )

    token_count = rate_limiter._count_tokens("test prompt")
    assert token_count == 10

    rate_limiter._stats["request_sizes"] = []
    token_count = rate_limiter._count_tokens(None)
    assert token_count == 100


def test_langchain_integration():
    """Test integration with LangChain."""
    tokenizer = Mock()
    tokenizer.return_value = 10

    rate_limiter = SmartRateLimiter(
        requests_per_minute=60,
        tokens_per_minute=4000,
        tokenizer=tokenizer,
        max_request_burst=3,
        max_token_burst=100,
        check_every_n_seconds=0.1,
    )

    # Create LangChain chat model with our rate limiter
    chat = ChatOpenAI(
        model_name="gpt-4",
        rate_limiter=rate_limiter,
        openai_api_key="fake-key",
    )

    def mock_generate(*args, **kwargs):
        time.sleep(0.1)
        return ChatResult(
            generations=[ChatGeneration(message=AIMessage(content="Test response"))]
        )

    # Patch _generate to let rate limiter execute
    with patch.object(chat, "_generate", side_effect=mock_generate) as mock_generate:
        # Process multiple requests
        prompts = ["Test prompt"] * 10
        with ThreadPoolExecutor(max_workers=3) as executor:
            results = list(executor.map(chat.invoke, prompts))
            assert all(isinstance(result.content, str) for result in results)

        # Verify results through rate limiter metrics
        metrics = rate_limiter.get_metrics()
        assert (
            metrics["total_requests_processed"] == 10
        ), "Should have processed 10 requests"
        assert (
            metrics["total_tokens_processed"] == 1000
        ), "Should have processed 1000 tokens"
        assert mock_generate.call_count == 10, "Should have called generate 10 times"


def test_langchain_integration_w_prompt_callback():
    """Test integration with LangChain."""
    tokenizer = Mock()
    tokenizer.return_value = 10

    rate_limiter = LangchainTokenAwareRateLimiter(
        requests_per_minute=60,
        tokens_per_minute=4000,
        tokenizer=tokenizer,
        max_request_burst=3,
        max_token_burst=100,
        check_every_n_seconds=0.1,
    )

    # Create LangChain chat model with our rate limiter
    chat = ChatOpenAI(
        model_name="gpt-4",
        rate_limiter=rate_limiter,
        openai_api_key="fake-key",
    )

    def mock_generate(*args, **kwargs):
        time.sleep(0.1)
        return ChatResult(
            generations=[ChatGeneration(message=AIMessage(content="Test response"))]
        )

    # Patch _generate to let rate limiter execute
    with patch.object(chat, "_generate", side_effect=mock_generate) as mock_generate:
        # Process multiple requests
        prompts = ["Test prompt"] * 10
        with ThreadPoolExecutor(max_workers=3) as executor:
            results = list(executor.map(chat.invoke, prompts))
            assert all(isinstance(result.content, str) for result in results)

        # Verify results through rate limiter metrics
        metrics = rate_limiter.get_metrics()
        assert (
            metrics["total_requests_processed"] == 10
        ), "Should have processed 10 requests"
        assert (
            metrics["total_tokens_processed"] == 1000
        ), "Should have processed 1000 tokens"
        assert mock_generate.call_count == 10, "Should have called generate 10 times"


@pytest.mark.asyncio
async def test_langchain_integration_w_prompt_callback_async():
    tokenizer = Mock()
    tokenizer.return_value = 10

    rate_limiter = LangchainTokenAwareRateLimiter(
        requests_per_minute=60,
        tokens_per_minute=4000,
        tokenizer=tokenizer,
        max_request_burst=3,
        max_token_burst=100,
        check_every_n_seconds=0.1,
    )
    chat = ChatOpenAI(
        model_name="gpt-4",
        rate_limiter=rate_limiter,
        openai_api_key="fake-key",
    )

    def mock_generate(*args, **kwargs):
        return ChatResult(
            generations=[ChatGeneration(message=AIMessage(content="Test response"))]
        )

    with patch.object(chat, "_agenerate", side_effect=mock_generate) as mock_generate:
        # Process multiple requests concurrently using asyncio
        prompts = ["Test prompt"] * 10
        tasks = [chat.ainvoke(prompt) for prompt in prompts]
        results = await asyncio.gather(*tasks)
        assert all(isinstance(result.content, str) for result in results)

        metrics = rate_limiter.get_metrics()
        assert (
            metrics["total_requests_processed"] == 10
        ), "Should have processed 10 requests"
        assert (
            metrics["total_tokens_processed"] == 1000
        ), "Should have processed 1000 tokens"
        assert mock_generate.call_count == 10, "Should have called generate 10 times"


@pytest.mark.skip(
    reason="This test simulates a large batch of requests, and thus take >1 minute to run. Recommend adhoc testing."
)
def test_langchain_integration_large_batch():
    """Test integration with LangChain for a large batch of requests."""
    tokenizer = Mock()
    tokenizer.return_value = 10

    rate_limiter = LangchainTokenAwareRateLimiter(
        requests_per_minute=60,  # 1 request/second
        tokens_per_minute=4000,  # ~67 tokens/second
        tokenizer=tokenizer,
        max_request_burst=3,  # Only 3 concurrent requests
        max_token_burst=100,  # Enough for concurrent requests
        check_every_n_seconds=0.1,
    )

    chat = ChatOpenAI(
        model_name="gpt-4",
        rate_limiter=rate_limiter,
        openai_api_key="fake-key",
    )

    def mock_generate(*args, **kwargs):
        # time.sleep(0.1)  # Simulate some processing time
        return ChatResult(
            generations=[ChatGeneration(message=AIMessage(content="Test response"))]
        )

    with patch.object(chat, "_generate", side_effect=mock_generate) as mock_generate:
        start_time = time.time()

        # Process 100 requests
        prompts = ["Test prompt"] * 100
        with ThreadPoolExecutor(max_workers=3) as executor:
            results = list(executor.map(chat.invoke, prompts))
            assert all(isinstance(result.content, str) for result in results)

        end_time = time.time()
        processing_time = end_time - start_time

        # Verify metrics
        metrics = rate_limiter.get_metrics()
        print(
            f"Processed {100/processing_time:.2f} requests per second; Time taken: {processing_time:.2f} seconds"
        )

        print(metrics)
        assert metrics["total_requests_processed"] == 100, "Should process all requests"
        assert metrics["total_tokens_processed"] == 1000, "Should process all tokens"
        assert mock_generate.call_count == 100, "Should make all calls"

        # Should take roughly 100 seconds (1 request/second)
        assert (
            95 <= processing_time <= 180
        ), f"Processing time was {processing_time} seconds"



import pytest
import json
import os
from unittest.mock import patch, MagicMock
from pydantic import BaseModel, Field
from langchain_llm_utils.request_cache import RequestCache
from langchain_llm_utils.llm import LLM, ModelProvider


class JokeResponse(BaseModel):
    setup: str = Field(default="What do you call a fish with no eyes?")
    punchline: str = Field(default="A fsh!")


@pytest.fixture
def mock_cache_file(tmp_path):
    """Fixture to create a temporary cache file. If the file already exists, it will be cleared."""
    cache_file = tmp_path / "test_cache.json"
    if os.path.exists(cache_file):
        os.remove(cache_file)
    return str(cache_file)


@pytest.fixture
def request_cache(mock_cache_file):
    """Fixture to create a RequestCache instance."""
    return RequestCache(cache_file=mock_cache_file)


@pytest.fixture
def mock_llm_provider():
    """Fixture to mock the LLM provider."""
    with patch(
        "langchain_llm_utils.llm.LLMFactory.create_provider"
    ) as mock_create_provider:
        mock_provider = MagicMock()
        mock_provider.with_structured_output.return_value = mock_provider
        mock_create_provider.return_value = mock_provider
        yield mock_provider


@pytest.fixture
def llm_instance_with_response_model(mock_llm_provider, request_cache):
    """Fixture to create an LLM instance with a response model for testing structured output."""
    return LLM(
        model_provider=ModelProvider.OPENAI,
        model_name="gpt-4o",
        response_model=JokeResponse,
        cache=request_cache,
    )


def test_cache_initialization(mock_cache_file):
    """Test cache initialization with new and existing cache files."""
    # Test with new cache file
    cache = RequestCache(cache_file=mock_cache_file)
    assert os.path.exists(mock_cache_file)


def test_cache_key_generation(request_cache):
    """Test cache key generation with different inputs."""
    key1 = request_cache.get_cache_key(
        model_name="gpt-4",
        model_provider=ModelProvider.OPENAI,
        final_prompt="test prompt",
    )
    key2 = request_cache.get_cache_key(
        model_name="gpt-4",
        model_provider=ModelProvider.OPENAI,
        final_prompt="test prompt",
    )
    key3 = request_cache.get_cache_key(
        model_name="gpt-4",
        model_provider=ModelProvider.OPENAI,
        final_prompt="different prompt",
    )

    # Same inputs should generate same key
    assert key1 == key2
    # Different prompts should generate different keys
    assert key1 != key3


def test_cache_set_get(request_cache):
    """Test basic cache set and get operations."""
    # Test with string value
    request_cache.set("key1", "value1")
    assert request_cache.get("key1") == "value1"

    # Test with dict value
    dict_value = {"nested": "value"}
    request_cache.set("key2", dict_value)
    assert request_cache.get("key2") == dict_value

    # Test non-existent key
    assert request_cache.get("nonexistent") is None


def test_cache_persistence(mock_cache_file):
    """Test that cache data persists between instances."""
    cache1 = RequestCache(cache_file=mock_cache_file)
    cache1.set("key1", "value1")
    cache1.save_cache(overwrite=True)

    # Create new instance pointing to same file
    cache2 = RequestCache(cache_file=mock_cache_file)
    assert cache2.get("key1") == "value1"


def test_cache_with_invalid_structured_output(request_cache):
    """Test cache with invalid structured output data."""
    invalid_data = {"invalid": "data"}
    request_cache.set("invalid_key", invalid_data)

    # Should return default model when data doesn't match schema
    result = request_cache.get_parsed("invalid_key", JokeResponse)
    assert isinstance(result, JokeResponse)
    assert result.setup == "What do you call a fish with no eyes?"


def test_generate_structured_output(
    llm_instance_with_response_model, mock_llm_provider, request_cache
):
    """Test the generate method with structured output."""
    # Mock the _llm.invoke method to return a structured response
    request_cache.clear()
    mock_response = JokeResponse()
    mock_llm_provider.invoke.return_value = mock_response

    response = llm_instance_with_response_model.generate("Tell me a joke")

    assert isinstance(response, JokeResponse)
    assert response.setup == mock_response.setup
    assert response.punchline == mock_response.punchline


def test_cache_with_llm_integration(
    mock_llm_provider, llm_instance_with_response_model, request_cache
):
    """Test cache integration with LLM class."""
    request_cache.clear()

    mock_response = JokeResponse()
    mock_llm_provider.invoke.return_value = mock_response

    # First call should hit the LLM since cache is empty
    result1 = llm_instance_with_response_model.generate("Tell me a joke")
    assert isinstance(result1, JokeResponse)
    assert result1.setup == mock_response.setup
    assert result1.punchline == mock_response.punchline
    assert mock_llm_provider.invoke.call_count == 1

    # Second call should hit cache
    result2 = llm_instance_with_response_model.generate("Tell me a joke")
    assert isinstance(result2, JokeResponse)
    assert result2.setup == mock_response.setup
    assert result2.punchline == mock_response.punchline
    assert mock_llm_provider.invoke.call_count == 1  # Shouldn't increase


def test_cache_clear(request_cache):
    """Test cache clearing functionality."""
    request_cache.set("key1", "value1")
    request_cache.set("key2", "value2")

    request_cache.clear()

    assert request_cache.get("key1") is None
    assert request_cache.get("key2") is None
    # Check if cache is empty rather than file size
    assert request_cache.cache == {}


def test_cache_with_template_variables(
    mock_llm_provider, llm_instance_with_response_model, request_cache
):
    """Test cache behavior with template variables in LLM."""
    request_cache.clear()
    mock_response = JokeResponse()
    mock_llm_provider.invoke.return_value = mock_response

    # Test with template and variables
    template = "Tell me a {type} joke about {topic}"
    variables = {"type": "programming", "topic": "arrays"}

    result1 = llm_instance_with_response_model.generate(
        template=template, input_variables=variables
    )
    assert mock_llm_provider.invoke.call_count == 1

    # Same template and variables should hit cache
    result2 = llm_instance_with_response_model.generate(
        template=template, input_variables=variables
    )
    assert mock_llm_provider.invoke.call_count == 1
    assert result1 == result2



import pytest
from unittest.mock import patch, MagicMock, AsyncMock
from langchain_llm_utils.evaluator import (
    ResponseLengthScore,
    CoherenceScore,
    Evaluator,
    EvaluatorConfig,
    BasicEvaluationSuite,
    ModelProvider,
    CoherenceResponse,
    EvaluatorCallback,
)
from langchain_core.messages import AIMessage
from pydantic import BaseModel, Field
from uuid import UUID
import time
import asyncio
from langchain_openai import ChatOpenAI
from langchain_core.outputs import ChatResult, ChatGeneration
from concurrent.futures import ThreadPoolExecutor


class TestResponse(BaseModel):
    text: str = Field(default_factory=lambda: "test")


@pytest.fixture
def mock_llm():
    """Fixture to mock the LLM instance in CoherenceScore."""
    with patch("langchain_llm_utils.evaluator.LLM") as mock:
        mock_instance = MagicMock()
        # Set up sync mock
        mock_instance.generate.return_value = CoherenceResponse(coherence=0.8)
        # Set up async mock
        async_mock = AsyncMock()
        async_mock.return_value = CoherenceResponse(coherence=0.8)
        mock_instance.agenerate = async_mock
        mock.return_value = mock_instance
        yield mock_instance


def test_response_length_score():
    """Test the ResponseLengthScore evaluator with different input types."""
    scorer = ResponseLengthScore()

    # Test with string input
    assert scorer.evaluate("prompt", "test response") == 13

    # Test with AIMessage input
    message = AIMessage(content="test response")
    assert scorer.evaluate("prompt", message) == 13

    # Test with BaseModel input
    model = TestResponse(text="test response")
    assert scorer.evaluate("prompt", model) == 24  # Length of JSON string


@pytest.mark.asyncio
async def test_response_length_score_async():
    """Test the async version of ResponseLengthScore."""
    scorer = ResponseLengthScore()
    score = await scorer.aevaluate("prompt", "test response")
    assert score == 13


def test_coherence_score(mock_llm):
    """Test the CoherenceScore evaluator."""
    scorer = CoherenceScore(llm_provider=ModelProvider.OPENAI, llm_name="gpt-4")

    score = scorer.evaluate("test prompt", "test response")
    assert score == 0.8

    # Verify the provider was called
    mock_llm.generate.assert_called_once()


@pytest.mark.asyncio
async def test_coherence_score_async(mock_llm):
    """Test the async version of CoherenceScore."""
    scorer = CoherenceScore(llm_provider=ModelProvider.OPENAI, llm_name="gpt-4")

    score = await scorer.aevaluate("test prompt", "test response")
    assert score == 0.8

    mock_llm.agenerate.assert_called_once()


def test_evaluator_config():
    """Test EvaluatorConfig configuration."""
    config = EvaluatorConfig()

    # Add heuristic scorer
    config.add_heuristic(ResponseLengthScore())
    assert len(config.scores) == 1

    # Add LLM judge scorer
    coherence_score = CoherenceScore(
        llm_provider=ModelProvider.OPENAI, llm_name="gpt-4"
    )
    config.add_llm_judge(coherence_score)
    assert len(config.scores) == 2


def test_evaluator(mock_llm):
    """Test the main Evaluator class."""
    config = EvaluatorConfig()
    config.add_heuristic(ResponseLengthScore())
    config.add_llm_judge(
        CoherenceScore(llm_provider=ModelProvider.OPENAI, llm_name="gpt-4")
    )

    evaluator = Evaluator(config)
    results = evaluator.evaluate("test prompt", "test response")

    assert len(results) == 2
    assert results[0].score == 13  # Length score
    assert results[1].score == 0.8  # Coherence score


@pytest.mark.asyncio
async def test_evaluator_async(mock_llm):
    """Test the async version of Evaluator."""
    config = EvaluatorConfig()
    config.add_heuristic(ResponseLengthScore())
    config.add_llm_judge(
        CoherenceScore(llm_provider=ModelProvider.OPENAI, llm_name="gpt-4")
    )

    evaluator = Evaluator(config)
    results = await evaluator.aevaluate("test prompt", "test response")

    assert len(results) == 2
    assert results[0].score == 13  # Length score
    assert results[1].score == 0.8  # Coherence score


def test_basic_evaluation_suite(mock_llm):
    """Test the BasicEvaluationSuite class."""
    suite = BasicEvaluationSuite()
    results = suite.evaluate("test prompt", "test response")

    assert len(results) == 2
    assert results[0].score == 13  # Length score
    assert results[1].score == 0.8  # Coherence score


@pytest.mark.asyncio
async def test_basic_evaluation_suite_async(mock_llm):
    """Test the async version of BasicEvaluationSuite."""
    suite = BasicEvaluationSuite()
    results = await suite.aevaluate("test prompt", "test response")

    assert len(results) == 2
    assert results[0].score == 13  # Length score
    assert results[1].score == 0.8  # Coherence score


def test_evaluator_error_handling(mock_llm):
    """Test error handling in Evaluator."""
    config = EvaluatorConfig()

    # Create a scorer that raises an exception
    faulty_scorer = ResponseLengthScore()
    faulty_scorer.evaluate = MagicMock(side_effect=Exception("Test error"))

    config.add_heuristic(faulty_scorer)
    config.add_heuristic(ResponseLengthScore())

    evaluator = Evaluator(config)
    results = evaluator.evaluate("test prompt", "test response")

    # Should still get results from the working scorer
    assert len(results) == 1
    assert results[0].score == 13


@pytest.mark.asyncio
async def test_evaluator_error_handling_async(mock_llm):
    """Test error handling in async Evaluator."""
    config = EvaluatorConfig()

    # Create a scorer that raises an exception
    faulty_scorer = ResponseLengthScore()
    faulty_scorer.aevaluate = MagicMock(side_effect=Exception("Test error"))

    config.add_heuristic(faulty_scorer)
    config.add_heuristic(ResponseLengthScore())

    evaluator = Evaluator(config)
    results = await evaluator.aevaluate("test prompt", "test response")

    # Should still get results from the working scorer
    assert len(results) == 1
    assert results[0].score == 13


@pytest.mark.parametrize(
    "sample_rate,expected_results",
    [
        (1.0, 2),  # Full sampling - should always evaluate
        (0.0, 0),  # No sampling - should never evaluate
    ],
)
def test_basic_evaluation_suite_sampling(mock_llm, sample_rate, expected_results):
    """Test that BasicEvaluationSuite respects sampling rates."""
    suite = BasicEvaluationSuite(sample_rate=sample_rate)
    results = suite.evaluate("test prompt", "test response")
    assert len(results) == expected_results


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "sample_rate,expected_results",
    [
        (1.0, 2),  # Full sampling - should always evaluate
        (0.0, 0),  # No sampling - should never evaluate
    ],
)
async def test_basic_evaluation_suite_sampling_async(
    mock_llm, sample_rate, expected_results
):
    """Test that async BasicEvaluationSuite respects sampling rates."""
    suite = BasicEvaluationSuite(sample_rate=sample_rate)
    results = await suite.aevaluate("test prompt", "test response")
    assert len(results) == expected_results


def test_evaluator_sampling_with_run_id(mock_llm):
    """Test that Evaluator sampling works consistently with run_ids."""
    # Create a fixed UUID for testing
    test_uuid = UUID("12345678-1234-5678-1234-567812345678")

    # Create evaluator with 50% sampling rate
    suite = BasicEvaluationSuite(sample_rate=0.5)

    # First evaluation with the UUID should determine if subsequent
    # evaluations with the same UUID are processed
    first_results = suite.evaluate("test prompt", "test response", run_id=test_uuid)
    expected_length = len(first_results)  # Will be either 0 or 2

    # Subsequent evaluations with same UUID should match first result
    second_results = suite.evaluate("test prompt", "test response", run_id=test_uuid)
    assert len(second_results) == expected_length


@pytest.mark.asyncio
async def test_evaluator_sampling_with_run_id_async(mock_llm):
    """Test that async Evaluator sampling works consistently with run_ids."""
    # Create a fixed UUID for testing
    test_uuid = UUID("12345678-1234-5678-1234-567812345678")

    # Create evaluator with 50% sampling rate
    suite = BasicEvaluationSuite(sample_rate=0.5)

    # First evaluation with the UUID should determine if subsequent
    # evaluations with the same UUID are processed
    first_results = await suite.aevaluate(
        "test prompt", "test response", run_id=test_uuid
    )
    expected_length = len(first_results)  # Will be either 0 or 2

    # Subsequent evaluations with same UUID should match first result
    second_results = await suite.aevaluate(
        "test prompt", "test response", run_id=test_uuid
    )
    assert len(second_results) == expected_length


def test_evaluator_langchain_integration(mock_llm):
    """Test Evaluator integration with LangChain."""

    # Create evaluator with full sampling
    suite = BasicEvaluationSuite(sample_rate=1.0)
    evaluator_callback = EvaluatorCallback(suite)

    # Create LangChain chat model with our evaluator callback
    chat = ChatOpenAI(
        model_name="gpt-4",
        callbacks=[evaluator_callback],
        openai_api_key="fake-key",
    )

    def mock_generate_result(*args, **kwargs):
        time.sleep(0.1)  # Simulate API call
        return ChatResult(
            generations=[ChatGeneration(message=AIMessage(content="Test response"))]
        )

    # We need to mock both the chat._generate and the coherence score evaluation
    with patch.object(
        chat, "_generate", side_effect=mock_generate_result
    ) as mock_generate, patch.object(
        mock_llm, "generate", return_value=CoherenceResponse(coherence=0.8)
    ):

        # Process multiple requests
        prompts = ["Test prompt"] * 10
        with ThreadPoolExecutor(max_workers=3) as executor:
            results = list(executor.map(chat.invoke, prompts))
            assert all(isinstance(result.content, str) for result in results)

        # Verify all requests were evaluated (since sample_rate=1.0)
        assert mock_generate.call_count == 10

        # Verify that evaluations were performed via callback
        # We need to add a way to track evaluations in the Evaluator class
        assert (
            len(suite.evaluation_results) == 10
        )  # This would need to be added to Evaluator

        # Verify the evaluation results
        for eval_result in suite.evaluation_results:
            assert len(eval_result) == 2  # Both scorers should have run
            assert eval_result[0].score_name == "ResponseLengthScore"
            assert eval_result[1].score_name == "CoherenceScore"
            assert eval_result[1].score == 0.8  # From our mocked coherence score


def test_evaluator_langchain_integration_with_sampling(mock_llm):
    """Test Evaluator integration with LangChain with partial sampling rate."""
    # Create evaluator with 70% sampling rate
    suite = BasicEvaluationSuite(sample_rate=0.7)
    evaluator_callback = EvaluatorCallback(suite)

    # Create LangChain chat model with our evaluator callback
    chat = ChatOpenAI(
        model_name="gpt-4",
        callbacks=[evaluator_callback],
        openai_api_key="fake-key",
    )

    def mock_generate_result(*args, **kwargs):
        time.sleep(0.1)  # Simulate API call
        return ChatResult(
            generations=[ChatGeneration(message=AIMessage(content="Test response"))]
        )

    # We need to mock both the chat._generate and the coherence score evaluation
    with patch.object(
        chat, "_generate", side_effect=mock_generate_result
    ) as mock_generate, patch.object(
        mock_llm, "generate", return_value=CoherenceResponse(coherence=0.8)
    ):
        # Process 100 requests to get a statistically significant sample
        num_requests = 100
        prompts = ["Test prompt"] * num_requests
        with ThreadPoolExecutor(max_workers=3) as executor:
            results = list(executor.map(chat.invoke, prompts))
            assert all(isinstance(result.content, str) for result in results)

        # Verify all requests were processed
        assert mock_generate.call_count == num_requests

        # Verify that evaluations were performed via callback
        # The number of evaluations should be approximately 70% of requests
        num_evaluations = len(suite.evaluation_results)
        expected_evaluations = int(num_requests * 0.7)
        margin_of_error = int(num_requests * 0.1)  # Allow 10% margin of error
        print("--------------------------------")
        print(f"Number of requests made to LLM: {mock_generate.call_count}")
        print(
            f"Number of evaluations performed: {num_evaluations}, "
            f"expected evaluations: {expected_evaluations}, "
            f"with margin of error +-{margin_of_error}"
        )
        assert abs(num_evaluations - expected_evaluations) <= margin_of_error, (
            f"Expected approximately {expected_evaluations} evaluations "
            f"(±{margin_of_error}), but got {num_evaluations}"
        )

        # Verify the evaluation results structure
        for eval_result in suite.evaluation_results:
            assert len(eval_result) == 2  # Both scorers should have run
            assert eval_result[0].score_name == "ResponseLengthScore"
            assert eval_result[1].score_name == "CoherenceScore"
            assert eval_result[1].score == 0.8  # From our mocked coherence score



import asyncio
import time
import threading
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Protocol
from langchain_core.rate_limiters import BaseRateLimiter
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.rate_limiters import InMemoryRateLimiter
from uuid import UUID
import tiktoken
from langchain_llm_utils.config import config

default_rate_limiter = InMemoryRateLimiter(
    requests_per_second=config.default_rate_limiter_requests_per_second,
    check_every_n_seconds=config.default_rate_limiter_check_every_n_seconds,
    max_bucket_size=config.default_rate_limiter_max_bucket_size,
)


class Tokenizer(Protocol):
    """Protocol for tokenizers that can count tokens in text."""

    def __call__(self, text: str) -> int:
        """Count tokens in the given text."""
        ...


class GPT4Tokenizer(Tokenizer):
    """Tokenizer for GPT-4 that can count tokens in text."""

    def __init__(self):
        self.tokenizer = tiktoken.encoding_for_model("gpt-4")

    def __call__(self, text: str) -> int:
        """Count tokens in the given text."""
        return len(self.enc.encode(text))


@dataclass
class TokenBucket:
    """
    Token bucket with timestamp for rate limiting.

    Attributes:
        tokens: Current number of available tokens
        last_update: Timestamp of last bucket update
        max_tokens: Maximum token capacity (burst limit)
    """

    tokens: float
    last_update: float
    max_tokens: float


class SmartRateLimiter(BaseRateLimiter):
    """
    Smart Rate Limiter for Token-Aware API Request Management

    This module implements an advanced rate limiter designed specifically for LLM API
    requests where both request count and token count need to be managed. It uses a
    dual token bucket algorithm combined with priority-based queuing to maximize
    throughput while maintaining strict rate limits.

    About token bucket algorithm: The bucket is filled with tokens at a given rate.
    Each request consumes a token. If there are not enough tokens in the bucket,
    the request is blocked until there are enough tokens.

    Algorithm Overview:
    ------------------
    1. Dual Token Bucket System
    - Request Bucket: Controls requests/minute
    - Token Bucket: Controls tokens/minute
    - Each bucket refills continuously at its specified rate
    - Burst capacity is configurable for both buckets

    2. Token Management
    - Integrated tokenizer for accurate token counting
    - Historical statistics for adaptive behavior
    - Fallback strategies for unknown token counts

    How it works:
    ----------------
    1. New Request → Count Tokens → Try Immediate Processing
    2. While Running:
        - Update Token Buckets
        - Sleep for Check Interval
    3. If (request_bucket.tokens ≥ 1 AND token_bucket.tokens ≥ required_tokens):
        - Consume Tokens
        - Return Success
    Else:
        - Return Failure

    Example Usage:
    -------------
    ```python
    # Using with GPT4Tokenizer (tiktoken)

    from utils.llm_utils import GPT4Tokenizer
    tokenizer = GPT4Tokenizer()

    limiter = SmartRateLimiter(
        requests_per_minute=60,
        tokens_per_minute=40_000,
        tokenizer=tokenizer
    )

    # Use with LangChain
    from langchain_openai import ChatOpenAI
    model = ChatOpenAI(
        model_name="gpt-4",
        rate_limiter=limiter
    )

    # Use with Lyra service
    from services.translation.models import GenericTranslationModel
    tm = GenericTranslationModel(
        name="ollama:llama3.1",
        rate_limiter=rate_limiter,
    )
    tm.predict(
        text="The capital of France is Paris.",
        target_language="French",
    )
    ```

    Example Scenario:
    ----------------
    - System configured for 60 requests/minute and 40,000 tokens/minute
    - Initial state: Both buckets full
    - We send requests in sequnce:
        t=0.0s: Request A (500 tokens)  → Processed immediately
        t=0.1s: Request B (800 tokens)  → Processed immediately
        t=0.2s: Request C (1000 tokens) → Queued (insufficient tokens)
        t=0.5s: Request D (300 tokens)  → Queued but higher priority than C
    - At t=1.0s: Buckets refill with 1s worth of capacity
        - D processed first (smaller size, decent wait time)
        - C processed when enough tokens accumulate
    """

    def __init__(
        self,
        requests_per_minute: float,
        tokens_per_minute: float,
        tokenizer: Tokenizer,
        max_request_burst: Optional[float] = None,
        max_token_burst: Optional[float] = None,
        check_every_n_seconds: float = 0.1,
        stats_window_seconds: float = 3600,  # 1 hour
    ):
        """
        Initialize the smart rate limiter.

        Args:
            requests_per_minute: Maximum requests per minute
            tokens_per_minute: Maximum tokens per minute
            tokenizer: Callable that counts tokens in text
            max_request_burst: Maximum burst size for requests
            max_token_burst: Maximum burst size for tokens
            check_every_n_seconds: Frequency to check if we can consume tokens
            stats_window_seconds: How long to keep historical stats
        """
        self.requests_per_second = requests_per_minute / 60.0
        self.tokens_per_second = tokens_per_minute / 60.0
        self.tokenizer = tokenizer
        self.check_every_n_seconds = check_every_n_seconds
        self.stats_window_seconds = stats_window_seconds
        self.callback: Optional[RateLimitCallback] = None

        # Initialize token buckets
        now = time.monotonic()
        self.request_bucket = TokenBucket(
            tokens=max_request_burst or requests_per_minute / 60.0,
            last_update=now,
            max_tokens=max_request_burst or requests_per_minute / 60.0,
        )
        self.token_bucket = TokenBucket(
            tokens=max_token_burst or tokens_per_minute / 60.0,
            last_update=now,
            max_tokens=max_token_burst or tokens_per_minute / 60.0,
        )

        # Thread safety
        self._lock = threading.Lock()

        # Stats with timestamps for rolling window
        self._stats = {
            "total_requests": 0,
            "total_tokens": 0,
            "request_sizes": [],  # List[Tuple[float, int]]  # (timestamp, size)
        }

    def _update_buckets(self) -> None:
        """
        Update both token buckets based on elapsed time.

        Implements a minimum update interval to prevent excessive updates.
        """
        now = time.monotonic()

        # Update request bucket
        elapsed = now - self.request_bucket.last_update
        self.request_bucket.tokens = min(
            self.request_bucket.max_tokens,
            self.request_bucket.tokens + (elapsed * self.requests_per_second),
        )
        self.request_bucket.last_update = now

        # Update token bucket
        elapsed = now - self.token_bucket.last_update
        self.token_bucket.tokens = min(
            self.token_bucket.max_tokens,
            self.token_bucket.tokens + (elapsed * self.tokens_per_second),
        )
        self.token_bucket.last_update = now

    def _can_consume(self, token_count: int) -> bool:
        """
        Check if we have capacity for both request and tokens.

        This method is atomic - if it returns True, it has already
        consumed the necessary tokens from both buckets.
        """
        with self._lock:
            self._update_buckets()

            if (
                self.request_bucket.tokens >= 1
                and self.token_bucket.tokens >= token_count
            ):
                self.request_bucket.tokens -= 1
                self.token_bucket.tokens -= token_count

                # Update stats
                self._stats["total_requests"] += 1
                self._stats["total_tokens"] += token_count
                return True

            # Calculate wait time based on token refill rates
            wait_for_request = (
                0
                if self.request_bucket.tokens >= 1
                else (1 - self.request_bucket.tokens) / self.requests_per_second
            )
            wait_for_tokens = (
                0
                if self.token_bucket.tokens >= token_count
                else (token_count - self.token_bucket.tokens) / self.tokens_per_second
            )

            # Return the longer wait time needed
            self._next_check_time = max(wait_for_request, wait_for_tokens)
            return False

    def _count_tokens(self, prompt: Optional[str] = None) -> int:
        """
        Count tokens in the prompt if available.

        Falls back to historical average if prompt is None.
        If there are no historical data, it default estimate of 100 tokens.
        """
        if prompt is None:
            # Use historical average or default
            recent_sizes = [size for _, size in self._stats["request_sizes"][-100:]]
            return int(sum(recent_sizes) / len(recent_sizes)) if recent_sizes else 100

        try:
            count = self.tokenizer(prompt)
            self._stats["request_sizes"].append((time.monotonic(), count))
            # print(f"Token count: {count}")
            return count
        except Exception:
            return 100  # Conservative fallback

    async def _aacquire(
        self, prompt: Optional[str] = None, *, blocking: bool = True
    ) -> bool:
        """
        Async acquire that handles token counting automatically.

        Args:
            prompt: Optional text to count tokens from
            blocking: Whether to wait for capacity

        Returns:
            True if capacity was acquired
        """
        token_count = self._count_tokens(prompt)

        if not blocking:
            return self._can_consume(token_count)

        while not self._can_consume(token_count):
            # Use the calculated wait time instead of fixed interval
            wait_time = min(self._next_check_time, self.check_every_n_seconds)
            await asyncio.sleep(wait_time)
        return True

    def _acquire(self, prompt: Optional[str] = None, *, blocking: bool = True) -> bool:
        """
        Synchronous version of acquire.

        Args:
            prompt: Optional text to count tokens from
            blocking: Whether to wait for capacity

        Returns:
            True if capacity was acquired
        """
        token_count = self._count_tokens(prompt)

        if self._can_consume(token_count):
            return True

        if not blocking:
            return False

        while not self._can_consume(token_count):
            # Use the calculated wait time instead of fixed interval
            wait_time = min(self._next_check_time, self.check_every_n_seconds)
            time.sleep(wait_time)

        return True

    async def aacquire(
        self,
        prompt: Optional[str] = None,
        *,
        blocking: bool = True,
        run_id: Optional[UUID] = None
    ) -> bool:
        """
        Async Get prompt from either direct arg or callback.

        Args:
            prompt: Direct prompt text (if available)
            blocking: Whether to wait for capacity
            run_id: The LangChain run ID to look up prompt
        """
        if prompt is None and self.callback and run_id:
            prompt = self.callback.get_prompt(run_id)
        return await self._aacquire(prompt=prompt, blocking=blocking)

    def acquire(
        self,
        prompt: Optional[str] = None,
        *,
        blocking: bool = True,
        run_id: Optional[UUID] = None
    ) -> bool:
        """
        Get prompt from either direct arg or callback.

        Args:
            prompt: Direct prompt text (if available)
            blocking: Whether to wait for capacity
            run_id: The LangChain run ID to look up prompt
        """

        # print(f"\nRate limiter acquire called with run_id: {run_id}")
        if prompt is None and self.callback and run_id:
            prompt = self.callback.get_prompt(run_id)
        # print(f"Prompt to use: {prompt}")
        return self._acquire(prompt=prompt, blocking=blocking)

    def register_callback(self, callback: "RateLimitCallback") -> None:
        """Register the callback that will provide prompts."""
        self.callback = callback

    def get_metrics(self) -> Dict[str, Any]:
        """
        Get current metrics about rate limiter state.

        Returns detailed statistics about usage patterns and current capacity.
        """
        with self._lock:
            self._update_buckets()

            metrics = {
                "available_requests": self.request_bucket.tokens,
                "available_tokens": self.token_bucket.tokens,
                "requests_per_minute": self.requests_per_second * 60,
                "tokens_per_minute": self.tokens_per_second * 60,
                "total_requests_processed": self._stats["total_requests"],
                "total_tokens_processed": self._stats["total_tokens"],
            }

            return metrics


class RateLimitWrapper(BaseRateLimiter):
    """Wrapper to bridge LangChain's simple acquire interface with our SmartRateLimiter."""

    def __init__(self, smart_limiter: SmartRateLimiter):
        self.smart_limiter = smart_limiter
        self._thread_local = threading.local()

    def _get_current_run_id(self) -> Optional[UUID]:
        return getattr(self._thread_local, "current_run_id", None)

    def set_current_run_id(self, run_id: UUID) -> None:
        self._thread_local.current_run_id = run_id

    def acquire(self, *, blocking: bool = True) -> bool:
        """Bridge method that adds run_id to the acquire call."""
        run_id = self._get_current_run_id()
        return self.smart_limiter.acquire(blocking=blocking, run_id=run_id)

    async def aacquire(self, *, blocking: bool = True) -> bool:
        """Async version of acquire."""
        run_id = self._get_current_run_id()
        return await self.smart_limiter.aacquire(blocking=blocking, run_id=run_id)


class RateLimitCallback(BaseCallbackHandler):
    """Callback handler that manages prompt tracking for rate limiting."""

    def __init__(self, wrapper: RateLimitWrapper):
        self.wrapper = wrapper
        self._thread_local = threading.local()

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        run_id: UUID,
        **kwargs: Any
    ) -> None:
        """Store prompt and set current run_id."""
        # print(f"\non_llm_start called with run_id: {run_id}")
        # print(f"Prompts received: {prompts}")
        if prompts:
            if not hasattr(self._thread_local, "prompts"):
                self._thread_local.prompts = {}
            self._thread_local.prompts[run_id] = prompts[0]
            # Set the current run_id in the wrapper
            self.wrapper.set_current_run_id(run_id)

    def get_prompt(self, run_id: UUID) -> Optional[str]:
        """Get prompt for current run_id and remove it from storage."""
        if not hasattr(self._thread_local, "prompts"):
            return None
        return self._thread_local.prompts.pop(run_id, None)


class LangchainTokenAwareRateLimiter(BaseRateLimiter):
    """
    Token-aware rate limiter for LangChain that manages both request and token rate limits.

    Example Usage:
    -------------
    ```python
    from utils.llm_utils import GPT4Tokenizer

    rate_limiter = LangchainTokenAwareRateLimiter(
        requests_per_minute=60,
        tokens_per_minute=40_000,
        tokenizer=GPT4Tokenizer(),
    )

    chat = ChatVertexAI(
        model="gemini-1.5-flash-001",
        rate_limiter=rate_limiter,
    )
    ```
    """

    def __init__(
        self,
        requests_per_minute: float,
        tokens_per_minute: float,
        tokenizer: Tokenizer,
        max_request_burst: Optional[float] = None,
        max_token_burst: Optional[float] = None,
        check_every_n_seconds: float = 0.1,
    ):
        """
        Initialize the rate limiter.

        Args:
            requests_per_minute: Maximum requests per minute
            tokens_per_minute: Maximum tokens per minute
            tokenizer: Callable that counts tokens in text
            max_request_burst: Maximum burst size for requests
            max_token_burst: Maximum burst size for tokens
            check_every_n_seconds: Frequency to check if we can consume tokens
        """
        # Create internal components
        self._smart_limiter = SmartRateLimiter(
            requests_per_minute=requests_per_minute,
            tokens_per_minute=tokens_per_minute,
            tokenizer=tokenizer,
            max_request_burst=max_request_burst,
            max_token_burst=max_token_burst,
            check_every_n_seconds=check_every_n_seconds,
        )
        self._wrapper = RateLimitWrapper(self._smart_limiter)
        self._callback = RateLimitCallback(self._wrapper)

        # Register callback with smart limiter
        self._smart_limiter.register_callback(self._callback)

        # Store for external access
        self.callbacks = [self._callback]

    def acquire(self, *, blocking: bool = True) -> bool:
        """Acquire rate limit permission."""
        return self._wrapper.acquire(blocking=blocking)

    async def aacquire(self, *, blocking: bool = True) -> bool:
        """Async acquire rate limit permission."""
        return await self._wrapper.aacquire(blocking=blocking)

    def get_metrics(self) -> Dict[str, Any]:
        """Get current metrics about rate limiter state."""
        return self._smart_limiter.get_metrics()



from typing import Type, Optional, Any, Union, Dict, Generic, List
from pydantic import BaseModel, Field, ValidationError
from langchain_core.prompts import PromptTemplate
from langchain_ollama import ChatOllama
from langchain_google_vertexai import ChatVertexAI
from langchain_openai import ChatOpenAI
from langchain_llm_utils.config import config
from langchain_llm_utils.common import get_logger, BaseModelType
from langchain_llm_utils.request_cache import RequestCache
from langchain_llm_utils.rate_limiter import default_rate_limiter
from httpx import ConnectError
from abc import ABC, abstractmethod
from enum import Enum
from langchain_core.messages import AIMessage, BaseMessage
from langfuse import Langfuse
from langfuse.decorators import observe, langfuse_context
import os
import asyncio

logger = get_logger("LLM")
langfuse_enabled = os.getenv("LANGFUSE_ENABLED", "false") == "true"

if not langfuse_enabled:
    os.environ["LANGFUSE_SECRET_KEY"] = ""
    os.environ["LANGFUSE_PUBLIC_KEY"] = ""
    os.environ["LANGFUSE_HOST"] = ""

langfuse = Langfuse(
    secret_key=os.getenv("LANGFUSE_SECRET_KEY"),
    public_key=os.getenv("LANGFUSE_PUBLIC_KEY"),
    host=os.getenv("LANGFUSE_HOST"),
    enabled=langfuse_enabled,
)


class ModelProvider(Enum):
    VERTEX_AI = "vertex_ai"
    OLLAMA = "ollama"
    OPENAI = "openai"


class ModelConfig(BaseModel):
    provider: ModelProvider
    base_model: str
    temperature: float = Field(ge=0.0, le=2.0)
    max_tokens: Optional[int] = None
    api_base: Optional[str] = None
    rate_limiter: Optional[Any] = default_rate_limiter
    structured_output_enabled: bool = True
    application_type: Optional[str] = "LLM"

    def validate_temperature(self):
        if not 0 <= self.temperature <= 2:
            raise ValueError("Temperature must be between 0 and 2")
        return True


class LLMProvider(ABC):
    @abstractmethod
    def invoke(self, prompt: str) -> BaseMessage:
        pass

    @abstractmethod
    async def ainvoke(self, prompt: str) -> BaseMessage:
        pass

    @abstractmethod
    def with_structured_output(
        self, response_model: Type[BaseModelType]
    ) -> "LLMProvider":
        pass


class OllamaProvider(LLMProvider):
    def __init__(self, model_config: ModelConfig):
        self.model_config = model_config
        self.client = ChatOllama(
            model=model_config.base_model,
            temperature=model_config.temperature,
            base_url=model_config.api_base,
            max_tokens=model_config.max_tokens,
            rate_limiter=model_config.rate_limiter,
        )

    def invoke(self, prompt: str) -> str:
        return self.client.invoke(prompt)

    async def ainvoke(self, prompt: str) -> str:
        return await self.client.ainvoke(prompt)

    def with_structured_output(
        self, response_model: Type[BaseModelType]
    ) -> "OllamaProvider":
        self.response_model = response_model
        if self.model_config.structured_output_enabled:
            logger.info(
                f"Setting structured output mode for {self.model_config.base_model} for {self.model_config.application_type}"
            )
            self.client = self.client.with_structured_output(response_model)
        return self


class VertexAIProvider(LLMProvider):
    def __init__(self, model_config: ModelConfig):
        self.model_config = model_config
        self.client = ChatVertexAI(
            model_name=model_config.base_model,
            project=config.GOOGLE_CLOUD_PROJECT,
            temperature=model_config.temperature,
            rate_limiter=model_config.rate_limiter,
        )

    def invoke(self, prompt: str) -> str:
        return self.client.invoke(prompt)

    async def ainvoke(self, prompt: str) -> str:
        return await self.client.ainvoke(prompt)

    def with_structured_output(
        self, response_model: Type[BaseModelType]
    ) -> "VertexAIProvider":
        self.response_model = response_model
        logger.info(
            f"Setting structured output mode for {self.model_config.base_model} for {self.model_config.application_type}"
        )
        self.client = self.client.with_structured_output(response_model)
        return self


class OpenAIProvider(LLMProvider):
    def __init__(self, model_config: ModelConfig):
        self.model_config = model_config
        self.client = ChatOpenAI(
            model=model_config.base_model,
            temperature=model_config.temperature,
            rate_limiter=model_config.rate_limiter,
            base_url=model_config.api_base,
        )

    def invoke(self, prompt: str) -> BaseMessage:
        return self.client.invoke(prompt)

    async def ainvoke(self, prompt: str) -> BaseMessage:
        return await self.client.ainvoke(prompt)

    def with_structured_output(
        self, response_model: Type[BaseModelType]
    ) -> "OpenAIProvider":
        self.response_model = response_model
        logger.info(
            f"Setting structured output mode for {self.model_config.base_model} for {self.model_config.application_type}"
        )
        self.client = self.client.with_structured_output(response_model)
        return self


class LLMFactory:
    """Initialize the appropriate LLM based on the model name.

    Currently supports:
    - Vertex AI (provided GOOGLE_CLOUD_PROJECT_ID and GOOGLE_APPLICATION_CREDENTIALS are available in the environment)
    - Ollama (provided OLLAMA_URL is available in the environment)
    - OpenAI (provided OPENAI_API_KEY is available in the environment)
    """

    _registry = {
        ModelProvider.OLLAMA: OllamaProvider,
        ModelProvider.OPENAI: OpenAIProvider,
        ModelProvider.VERTEX_AI: VertexAIProvider,
    }

    @classmethod
    def create_config(
        cls,
        model_provider: ModelProvider,
        model_name: str,
        temperature: float = 0.0,
        max_tokens: int = 256,
        rate_limiter: Any = default_rate_limiter,
        application_type: Optional[str] = None,
    ) -> ModelConfig:
        """Create model config from model string.

        Example model strings:
        - gemini-1.5-flash
        - ollama:llama3.1
        - ollama:llama3.1:70b
        - openai:gpt-4o

        Custom gguf models on Ollama are not supported for structured output.
        """
        custom_gguf_models = config.CUSTOM_GGUF_MODELS
        if model_provider == ModelProvider.VERTEX_AI:
            return ModelConfig(
                provider=ModelProvider.VERTEX_AI,
                base_model=model_name,
                temperature=temperature,
                rate_limiter=rate_limiter,
                application_type=application_type,
            )
        elif model_provider == ModelProvider.OLLAMA:
            base_model = model_name
            return ModelConfig(
                provider=ModelProvider.OLLAMA,
                base_model=base_model,
                temperature=temperature,
                max_tokens=max_tokens,
                api_base=config.OLLAMA_URL,
                rate_limiter=rate_limiter,
                structured_output_enabled=(
                    False if base_model in custom_gguf_models else True
                ),
                application_type=application_type,
            )
        elif model_provider == ModelProvider.OPENAI:
            return ModelConfig(
                provider=ModelProvider.OPENAI,
                base_model=model_name,
                temperature=temperature,
                rate_limiter=rate_limiter,
                api_base=config.OPENAI_API_BASE,
                application_type=application_type,
            )
        else:
            raise ValueError(f"Unsupported model provider: {model_provider}")

    @classmethod
    def create_provider(cls, model_config: ModelConfig) -> LLMProvider:
        """Create LLM provider instance based on config."""
        provider_class = cls._registry.get(model_config.provider)
        if not provider_class:
            raise ValueError(f"Unsupported provider: {model_config.provider}")
        return provider_class(model_config)


class LLM(Generic[BaseModelType]):
    """
    A unified interface for interacting with different LLM implementations.
    Supports both structured and unstructured outputs, caching, and temperature control.

    This class provides a simple way to:
    1. Generate responses from different LLM providers (Vertex AI, Ollama)
    2. Handle both structured (Pydantic models) and unstructured (string) outputs
    3. Cache responses for faster repeated queries
    4. Control temperature and other model parameters

    Example usage:

    ```python
    # Simple string output
    llm = LLM(model_provider='vertex-ai', model_name="gemini-1.5-flash", temperature=0.2)
    response = llm.generate("Tell me a joke")

    # With template
    response = llm.generate(
        template="Tell me a {type} joke about {topic}",
        input_variables={"type": "dad", "topic": "programming"}
    )

    # Structured output with caching
    from pydantic import BaseModel

    class JokeResponse(BaseModel):
        setup: str
        punchline: str

    cache = RequestCache("jokes_cache.json")
    llm = LLM[JokeResponse](
        model="gemini-1.5-flash",
        cache=cache,
        response_model=JokeResponse
    )

    joke = llm.generate("Tell me a joke")
    print(f"{joke.setup} - {joke.punchline}")
    ```
    """

    def __init__(
        self,
        model_provider: ModelProvider,
        model_name: str = None,
        temperature: float = 0.0,
        max_tokens: int = 256,
        cache: Optional[RequestCache] = None,
        response_model: Optional[Type[BaseModelType]] = None,
        model_type: Optional[str] = "invoke",
        rate_limiter: Optional[Any] = None,
        evaluator: Optional["Evaluator"] = None,  # type: ignore
        langfuse_tags: Optional[List[str]] = None,
    ):
        self.model_provider = model_provider
        self.model_name = model_name
        self.config = LLMFactory.create_config(
            model_provider=self.model_provider,
            model_name=self.model_name,
            temperature=temperature,
            max_tokens=max_tokens,
            rate_limiter=rate_limiter,
            application_type=model_type,
        )
        self.cache = cache
        self.response_model = response_model
        self.model_type = model_type
        self.evaluator = evaluator
        self.langfuse_tags = langfuse_tags
        self._llm = self._initialize_llm()

        if response_model and self._is_valid_response_model(response_model):
            self._llm = self._llm.with_structured_output(response_model)

    def _is_valid_response_model(self, response_model: Type[BaseModelType]) -> bool:
        """Check if the response model is valid by validating that all fields have default values."""
        try:
            return isinstance(response_model(), response_model)
        except Exception as e:
            logger.error(
                f"Error: {e} Make sure all fields in {response_model} have default values."
            )
            return False

    def _initialize_llm(self) -> LLMProvider:
        """Initialize the appropriate LLM provider based on config."""
        try:
            provider = LLMFactory.create_provider(self.config)
            return provider
        except Exception as e:
            logger.error(f"Failed to initialize LLM provider: {e}")
            raise

    def _format_prompt(
        self,
        prompt: str,
        template: Optional[str] = None,
        input_variables: Optional[Dict[str, Any]] = None,
    ) -> str:
        """Format the prompt using template if provided."""
        if template and input_variables:
            prompt_template = PromptTemplate(
                template=template, input_variables=list(input_variables.keys())
            )
            return prompt_template.format(**input_variables)
        return prompt

    @observe(as_type="generation")
    def invoke(self, prompt: str) -> Union[str, BaseModelType]:
        """Invoke the LLM with a prompt."""
        try:
            res = self._llm.invoke(prompt)

            # Store the current observation ID
            observation_id = langfuse_context.get_current_observation_id()
            trace_id = langfuse_context.get_current_trace_id()
            logger.debug(
                f"Current observation ID: {observation_id}, trace ID: {trace_id}, model name: {self.model_name}, model type: {self.model_type}"
            )
            langfuse_context.update_current_observation(
                input=prompt,
                model=self.model_name,
                output=res,
                name=self.model_type,
            )

            # Run evaluations if evaluator exists
            if self.evaluator:
                self._run_evaluations(
                    prompt, res, trace_id, observation_id, run_id=trace_id
                )

            return res
        except (ConnectionRefusedError, ConnectionError, ConnectError) as e:
            logger.error(f"Connection refused invoking LLM {self.model_name}: {e}")
            raise type(e)(
                f"Failed to connect to LLM {self.model_name}. Please check if the service is running and accessible."
            )
        except Exception as e:
            logger.error(f"Error invoking LLM {self.model_name}: {e}")
            return None

    @observe(as_type="generation")
    async def ainvoke(self, prompt: str) -> Union[str, BaseModelType]:
        """Async version of invoke method."""
        try:
            res = await self._llm.ainvoke(prompt)

            # Store the current observation ID
            observation_id = langfuse_context.get_current_observation_id()
            trace_id = langfuse_context.get_current_trace_id()
            logger.debug(
                f"AInvoke Current observation ID: {observation_id}, trace ID: {trace_id}, model name: {self.model_name}, model type: {self.model_type}"
            )
            langfuse_context.update_current_observation(
                input=prompt,
                model=self.model_name,
                output=res,
                name=self.model_type,
            )

            # Run evaluations if evaluator exists
            if self.evaluator:
                asyncio.create_task(
                    self._arun_evaluations(
                        prompt, res, trace_id, observation_id, run_id=trace_id
                    )
                )

            return res
        except (ConnectionRefusedError, ConnectionError, ConnectError) as e:
            logger.error(f"Connection refused invoking LLM {self.model_name}: {e}")
            raise type(e)(
                f"Failed to connect to LLM {self.model_name}. Please check if the service is running and accessible."
            )
        except Exception as e:
            logger.error(f"Error invoking LLM {self.model_name}: {e}")
            return None

    def _run_evaluations(
        self,
        prompt: str,
        response: Union[str, BaseModelType],
        trace_id: str,
        observation_id: str,
        run_id: str,
    ):
        """Run evaluations synchronously and update Langfuse scores."""
        try:
            logger.debug(
                f"Running evaluations for trace {trace_id} and observation {observation_id}"
            )
            evaluations = self.evaluator.evaluate(prompt, response, run_id)
            logger.debug(
                f"Pushing evaluations to Langfuse for trace {trace_id} and observation {observation_id}"
            )
            for eval in evaluations:
                langfuse.score(
                    trace_id=trace_id,
                    observation_id=observation_id,
                    value=eval.score,
                    name=eval.score_name,
                    comment=eval.details["description"] if eval.details else "",
                )
        except Exception as e:
            logger.error(f"Error running evaluations in LLM: {e}")

    async def _arun_evaluations(
        self,
        prompt: str,
        response: Union[str, BaseModelType],
        trace_id: str,
        observation_id: str,
        run_id: str,
    ):
        """Run evaluations asynchronously and update Langfuse scores."""
        try:
            logger.debug(
                f"Running evaluations for trace {trace_id} and observation {observation_id}"
            )
            evaluations = await self.evaluator.aevaluate(prompt, response, run_id)
            logger.debug(
                f"Pushing evaluations to Langfuse for trace {trace_id} and observation {observation_id}"
            )
            for eval in evaluations:
                langfuse.score(
                    trace_id=trace_id,
                    observation_id=observation_id,
                    value=eval.score,
                    name=eval.score_name,
                    comment=eval.details["description"] if eval.details else "",
                )
        except Exception as e:
            logger.error(f"Error running evaluations in LLM: {e}")

    def _handle_cached_response(
        self,
        final_prompt: str,
    ) -> Optional[BaseModelType]:
        """Handle cached response retrieval and storage."""
        cache_key = self.cache.get_cache_key(
            model_name=self.model_name,
            model_provider=self.model_provider,
            final_prompt=final_prompt,
        )
        # Try to get from cache
        return self.cache.get_parsed(cache_key, self.response_model)

    def _cache_response(self, final_prompt: str, response: Any) -> None:
        """Cache the response if it exists."""
        if response:
            cache_key = self.cache.get_cache_key(
                model_name=self.model_name,
                model_provider=self.model_provider,
                final_prompt=final_prompt,
            )
            self.cache.set(cache_key, response)

    def _process_structured_response(self, res: Any) -> Union[str, BaseModelType]:
        """Process and validate structured response."""
        if not self.response_model:
            return res

        # Return empty response model if no result
        if not res:
            return self.response_model()

        # Handle structured output validation
        if not isinstance(res, self.response_model):
            try:
                if isinstance(res, AIMessage):
                    res = res.content
                res = self.response_model.model_validate_json(res)
            except ValidationError as e:
                logger.error(f"Failed to validate structured output: {e}")
                return self.response_model()

        return res

    @observe()
    def generate(
        self,
        prompt: str = "",
        template: Optional[str] = None,
        input_variables: Optional[Dict[str, Any]] = None,
    ) -> Union[str, BaseModelType]:
        """
        Generate a response for a prompt or generate a response from a template.

        Args:
            prompt: Direct prompt string if template is None
            template: Optional template string with placeholders
            input_variables: Dict of variables to fill template placeholders

        Returns:
            Either a string response or an instance of response_model if specified
        """
        final_prompt = self._format_prompt(prompt, template, input_variables)

        langfuse_context.update_current_observation(
            model=self.model_name,
            name=self.model_type,
            tags=self.langfuse_tags,
        )

        # Handle cached responses if caching is enabled
        if self.response_model and self.cache:
            cached_response = self._handle_cached_response(final_prompt)
            if cached_response is not None:
                return cached_response

            response = self.invoke(final_prompt)
            self._cache_response(final_prompt, response)
            return response

        # Generate and process response
        res = self.invoke(final_prompt)
        return self._process_structured_response(res)

    @observe()
    async def agenerate(
        self,
        prompt: str = "",
        template: Optional[str] = None,
        input_variables: Optional[Dict[str, Any]] = None,
    ) -> Union[str, BaseModelType]:
        """
        Async version of generate method.
        """
        final_prompt = self._format_prompt(prompt, template, input_variables)
        langfuse_context.update_current_observation(
            model=self.model_name,
            name=self.model_type,
            tags=self.langfuse_tags,
        )

        # Handle cached responses if caching is enabled
        if self.response_model and self.cache:
            cached_response = self._handle_cached_response(final_prompt)
            if cached_response is not None:
                return cached_response

            response = await self.ainvoke(final_prompt)
            self._cache_response(final_prompt, response)
            return response

        # Generate and process response
        res = await self.ainvoke(final_prompt)
        return self._process_structured_response(res)



from typing import Optional, List, Dict, Any, Protocol, Type, Union
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from langchain_llm_utils.llm import LLM, ModelProvider
from langchain_llm_utils.common import BaseModelType, get_logger, decider
from langchain_core.messages import AIMessage
from langchain_core.callbacks import BaseCallbackHandler
from pydantic import BaseModel, Field
import asyncio
import threading
from uuid import UUID

logger = get_logger("Evaluator")


class EvaluationType(Enum):
    HEURISTIC = "heuristic"
    LLM_JUDGE = "llm_judge"


class EvaluationScore(Protocol):

    @classmethod
    def get_string_response(cls, response: Union[str, BaseModel, AIMessage]) -> str:
        if isinstance(response, AIMessage):
            response = response.content
        elif isinstance(response, BaseModel):
            response = response.model_dump_json()
        return response

    def evaluate(self, prompt: str, response: str) -> float:
        """Return a score between 0 and 1"""
        ...

    async def aevaluate(self, prompt: str, response: str) -> float:
        """Return a score between 0 and 1"""
        ...


@dataclass(frozen=True)
class EvaluationResult:
    score: float
    score_name: str
    details: Dict[str, Any] = None


class BaseHeuristicScore(EvaluationScore):
    def __init__(self, description: str):
        self.description = description

    @abstractmethod
    def evaluate(self, prompt: str, response: str) -> float:
        pass

    @abstractmethod
    async def aevaluate(self, prompt: str, response: str) -> float:
        pass


class BaseLLMJudgeScore(EvaluationScore):
    def __init__(
        self,
        llm_provider: ModelProvider,
        llm_name: str,
        response_model: Type[BaseModelType],
        description: str,
        llm_judge_type: str = "llm_judge",
    ):
        self.llm_provider = llm_provider
        self.llm_name = llm_name
        self.response_model = response_model
        self.description = description
        self.llm_judge = LLM(
            model_provider=self.llm_provider,
            model_name=self.llm_name,
            response_model=self.response_model,
            model_type=llm_judge_type,
            langfuse_tags=["llm_judge", llm_judge_type],
        )

    @abstractmethod
    def evaluate(self, prompt: str, response: str) -> float:
        pass

    @abstractmethod
    async def aevaluate(self, prompt: str, response: str) -> float:
        pass


class ResponseLengthScore(BaseHeuristicScore):
    """
    Length of the response.
    """

    def __init__(self):
        description = "Length of the response"
        super().__init__(description)

    def evaluate(
        self, prompt: str, response: Union[str, BaseModel, AIMessage]
    ) -> float:
        response = self.get_string_response(response)
        return len(response)

    async def aevaluate(
        self, prompt: str, response: Union[str, BaseModel, AIMessage]
    ) -> float:
        response = self.get_string_response(response)
        return len(response)


class CoherenceResponse(BaseModel):
    coherence: float = Field(
        description="The coherence of the response", default_factory=lambda: 0.0
    )


class CoherenceScore(BaseLLMJudgeScore):

    def __init__(self, llm_provider: ModelProvider, llm_name: str):
        description = "Score the coherence of the response given the prompt"
        super().__init__(
            llm_provider,
            llm_name,
            response_model=CoherenceResponse,
            description=description,
            llm_judge_type="coherence_judge",
        )
        self.evaluation_judge_prompt = """
        Rate the coherence of this response to the prompt on a scale of 0-1.
        Return in JSON format:
        {{
            "coherence": "score (0-1)"
        }}
        
        Prompt: {prompt}
        Response: {response}
        
        Score: 
        """

    def evaluate(
        self, prompt: str, response: Union[str, BaseModel, AIMessage]
    ) -> float:
        response = self.get_string_response(response)
        result = self.llm_judge.generate(
            template=self.evaluation_judge_prompt,
            input_variables={"prompt": prompt, "response": response},
        )

        return float(result.coherence)

    async def aevaluate(
        self, prompt: str, response: Union[str, BaseModel, AIMessage]
    ) -> float:
        response = self.get_string_response(response)
        result = await self.llm_judge.agenerate(
            template=self.evaluation_judge_prompt,
            input_variables={"prompt": prompt, "response": response},
        )

        return float(result.coherence)


class EvaluatorConfig:
    def __init__(self):
        self.scores: List[EvaluationScore] = []

    def add_heuristic(self, score: BaseHeuristicScore) -> "EvaluatorConfig":
        self.scores.append(score)
        return self

    def add_llm_judge(self, score: BaseLLMJudgeScore) -> "EvaluatorConfig":
        self.scores.append(score)
        return self


class Evaluator:
    def __init__(
        self, config: Optional[EvaluatorConfig] = None, sample_rate: float = 1.0
    ):
        self.config = config or EvaluatorConfig()
        self.sample_rate = sample_rate
        self.evaluation_results = []  # Store evaluation results

    def _should_evaluate(self, run_id: Optional[UUID] = None) -> bool:
        """Determine if evaluation should be performed based on sampling."""
        if self.sample_rate >= 1.0:
            return True
        if self.sample_rate <= 0.0:
            return False
        if run_id is None:
            return True  # Always evaluate if no run_id provided
        should_evaluate = decider(run_id, self.sample_rate, "evaluator")
        logger.debug(
            f"Input run_id: {run_id}, should evaluate: {should_evaluate}, sample_rate: {self.sample_rate}"
        )
        return should_evaluate

    def evaluate(
        self, prompt: str, response: str, run_id: Optional[UUID] = None
    ) -> List[EvaluationResult]:
        """Evaluate with optional sampling based on run_id."""

        # Check if we should evaluate based on sampling
        if not self._should_evaluate(run_id):
            return []

        results = []
        for score in self.config.scores:
            try:
                score_value = score.evaluate(prompt, response)
                result = EvaluationResult(
                    score=score_value,
                    score_name=score.__class__.__name__,
                    details={"description": score.description},
                )
                results.append(result)
            except Exception as e:
                logger.error(
                    f"Error evaluating score {score.__class__.__name__}: {e}",
                    exc_info=True,
                )

        # Store the results
        self.evaluation_results.append(results)
        return results

    async def aevaluate(
        self, prompt: str, response: str, run_id: Optional[UUID] = None
    ) -> List[EvaluationResult]:
        """Async evaluate with optional sampling based on run_id."""

        # Check if we should evaluate based on sampling
        if not self._should_evaluate(run_id):
            return []

        results = []
        for score in self.config.scores:
            try:
                result = await score.aevaluate(prompt, response)
                results.append(
                    EvaluationResult(
                        score=result,
                        score_name=score.__class__.__name__,
                        details={"description": score.description},
                    )
                )
            except Exception as e:
                logger.error(
                    f"Error evaluating score {score.__class__.__name__}: {e}",
                    exc_info=True,
                )
        return results


class EvaluatorCallback(BaseCallbackHandler):
    """Callback handler that tracks run IDs for evaluation sampling."""

    def __init__(self, evaluator: Evaluator):
        self._thread_local = threading.local()
        self.evaluator = evaluator

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        run_id: UUID,
        **kwargs: Any,
    ) -> None:
        """Store the current run_id and prompt."""
        if not hasattr(self._thread_local, "current_run_id"):
            self._thread_local.current_run_id = None
        self._thread_local.current_run_id = run_id
        self._thread_local.current_prompt = prompts[0] if prompts else None

    def on_llm_end(
        self,
        response: Any,
        run_id: UUID,
        **kwargs: Any,
    ) -> None:
        """Trigger evaluation when LLM response is received."""
        prompt = getattr(self._thread_local, "current_prompt", None)
        if prompt and hasattr(response, "generations"):
            # Trigger evaluation
            self.evaluator.evaluate(prompt, response, run_id=run_id)


class BasicEvaluationSuite(Evaluator):
    def __init__(self, sample_rate: float = 1.0):
        config = EvaluatorConfig()
        config.add_heuristic(ResponseLengthScore())
        config.add_llm_judge(
            CoherenceScore(
                llm_provider=ModelProvider.OPENAI,
                llm_name="gpt-4o",
            )
        )
        super().__init__(config=config, sample_rate=sample_rate)



from typing import List, Optional, Callable
from enum import Enum
from dataclasses import dataclass
import multiprocessing
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm
from langchain_llm_utils.common import T, R
from langchain_llm_utils.common import get_logger

logger = get_logger("BatchProcessor")


class WorkloadType(Enum):
    CPU_BOUND = "cpu_bound"
    IO_BOUND = "io_bound"
    MIXED = "mixed"


@dataclass
class BatchProcessConfig:
    workload_type: WorkloadType
    cpu_allocation: float = 0.75
    min_workers: int = 1
    max_workers: Optional[int] = None
    description: str = "Processing"
    unit: str = "item"


def batch_process_with_progress(
    items: List[T],
    process_func: Callable[[T], R],
    config: BatchProcessConfig,
) -> List[R]:
    """
    Generic utility for batch processing items with real-time progress tracking.

    Args:
        items: List of items to process
        process_func: Function to process each item
        config: BatchProcessConfig object containing processing parameters
                config.workload_type :
                    If config.workload_type is CPU_BOUND, then the number of workers will be
                    the number of CPUs minus 1.
                    If config.workload_type is IO_BOUND, then the number of workers will be
                    twice the number of CPUs.
                    If config.workload_type is MIXED, then the number of workers will be
                    the number of CPUs times the cpu_allocation.
                config.max_workers :
                    If config.max_workers is not None, then the number of workers will be
                    the minimum of the number of workers calculated above and config.max_workers.

    Returns:
        List of processed results
    """
    cpu_count = multiprocessing.cpu_count()

    # Calculate optimal workers
    if config.workload_type == WorkloadType.CPU_BOUND:
        optimal_workers = max(config.min_workers, cpu_count - 1)
    elif config.workload_type == WorkloadType.IO_BOUND:
        optimal_workers = cpu_count * 2
    else:  # MIXED
        optimal_workers = max(
            config.min_workers, int(cpu_count * config.cpu_allocation)
        )

    if config.max_workers:
        optimal_workers = min(optimal_workers, config.max_workers)

    # Initialize results list with None values to maintain order
    results = [None] * len(items)

    with ThreadPoolExecutor(max_workers=optimal_workers) as executor:
        # Submit all tasks and store futures with their indices
        future_to_index = {
            executor.submit(process_func, item): idx for idx, item in enumerate(items)
        }

        # Create progress bar
        with tqdm(total=len(items), desc=config.description, unit=config.unit) as pbar:
            # Process completed futures as they finish
            for future in as_completed(future_to_index):
                idx = future_to_index[future]
                try:
                    result = future.result()
                    results[idx] = result
                except Exception as e:
                    results[idx] = e
                pbar.update(1)

    # Check for exceptions
    for i, result in enumerate(results):
        if isinstance(result, Exception):
            raise result
    return results



from pydantic_settings import BaseSettings
import os
from typing import List

custom_gguf_models = "rr_model_v2"


class Config(BaseSettings):
    GOOGLE_CLOUD_PROJECT_ID: str = os.getenv("GOOGLE_CLOUD_PROJECT_ID", "")
    OLLAMA_URL: str = os.getenv("OLLAMA_URL", "http://localhost:11434")
    CUSTOM_GGUF_MODELS: List[str] = os.getenv(
        "CUSTOM_GGUF_MODELS", custom_gguf_models
    ).split(",")
    OPENAI_API_BASE: str = os.getenv("OPENAI_API_BASE", "https://api.openai.com/v1")
    default_rate_limiter_requests_per_second: float = os.getenv(
        "DEFAULT_RATE_LIMITER_REQUESTS_PER_SECOND", 10
    )
    default_rate_limiter_check_every_n_seconds: float = os.getenv(
        "DEFAULT_RATE_LIMITER_CHECK_EVERY_N_SECONDS", 0.5
    )
    default_rate_limiter_max_bucket_size: int = os.getenv(
        "DEFAULT_RATE_LIMITER_MAX_BUCKET_SIZE", 10
    )


config = Config()



import pytest
from typing import List, Dict
import asyncio
import time
import tiktoken
from dataclasses import dataclass
from concurrent.futures import ThreadPoolExecutor
import threading
from langchain_core.rate_limiters import InMemoryRateLimiter
from langchain_llm_utils.rate_limiter import default_rate_limiter, SmartRateLimiter


@dataclass
class MockLLMResponse:
    text: str
    tokens_used: int
    request_time: float


class MockLLMAPI:
    """Simulates an LLM API with rate limits and processing delays"""

    def __init__(self, rpm_limit: int, tpm_limit: int):
        self.rpm_limit = rpm_limit
        self.tpm_limit = tpm_limit
        self._request_times = []
        self._token_times = []
        self._lock = threading.Lock()
        self.total_requests = 0
        self.total_tokens = 0
        self.rejected_requests = 0

    def _check_rate_limits(self, tokens: int) -> bool:
        """Check if the request would exceed rate limits"""
        now = time.time()
        minute_ago = now - 60

        with self._lock:
            # Clean up old entries
            self._request_times = [t for t in self._request_times if t > minute_ago]
            self._token_times = [t for t in self._token_times if t > minute_ago]

            # Check limits
            if len(self._request_times) >= self.rpm_limit:
                return False
            if sum(self._token_times) + tokens >= self.tpm_limit:
                return False

            # Update tracking
            self._request_times.append(now)
            self._token_times.append(tokens)
            self.total_requests += 1
            self.total_tokens += tokens
            return True

    def process_prompt(self, prompt: str) -> MockLLMResponse:
        """Process a prompt, respecting rate limits

        We first check if the request would exceed rate limits. If it does, we raise an exception.
        If it doesn't, we simulate processing time which is linear with the number of tokens in the prompt.

        """
        enc = tiktoken.encoding_for_model("gpt-4")
        tokens = len(enc.encode(prompt))

        if not self._check_rate_limits(tokens):
            self.rejected_requests += 1
            raise Exception("Rate limit exceeded")

        # Simulate processing time (longer for longer prompts)
        process_time = 0.1 + (
            tokens * 0.01
        )  # Base latency + token-based processing time
        time.sleep(process_time)

        return MockLLMResponse(
            text=f"Processed: {prompt[:10]}...",
            tokens_used=tokens,
            request_time=process_time,
        )

    async def aprocess_prompt(self, prompt: str) -> MockLLMResponse:
        """Async version of process_prompt"""
        enc = tiktoken.encoding_for_model("gpt-4")
        tokens = len(enc.encode(prompt))

        if not self._check_rate_limits(tokens):
            self.rejected_requests += 1
            raise Exception("Rate limit exceeded")

        # Simulate processing time (longer for longer prompts)
        process_time = 0.1 + (
            tokens * 0.01
        )  # Base latency + token-based processing time
        await asyncio.sleep(process_time)  # Use asyncio.sleep instead of time.sleep

        return MockLLMResponse(
            text=f"Processed: {prompt[:10]}...",
            tokens_used=tokens,
            request_time=process_time,
        )


def print_test_metrics(
    label: str,
    duration: float,
    successful_results: int,
    total_prompts: int,
    api_stats: MockLLMAPI,
    successful_tokens: int,
):
    """Print metrics for a single test run"""
    print(f"\nTesting with {label}:")
    print(f"Duration: {duration:.2f}s")
    print(f"Successful requests: {successful_results}/{total_prompts}")
    print(f"Success rate: {(successful_results/total_prompts)*100:.1f}%")
    print(f"Rejected requests: {api_stats.rejected_requests}")
    print(f"Attempted requests: {api_stats.total_requests}")
    print(f"Successful RPM: {successful_results / (duration / 60):.2f}")
    print(f"Successful tokens processed: {successful_tokens}")
    print(f"Successful TPM: {successful_tokens / (duration / 60):.2f}")
    print(f"Attempted RPM: {api_stats.total_requests / (duration / 60):.2f}")
    print(f"Attempted TPM: {api_stats.total_tokens / (duration / 60):.2f}")


def print_comparison_metrics(
    baseline_stats: Dict,
    smart_stats: Dict,
    langchain_stats: Dict,
    total_prompts: int,
    smart_limiter_metrics: Dict,
    is_async: bool = False,
):
    """Print comparison metrics between different rate limiters"""
    async_label = "(async)" if is_async else ""

    # Print smart limiter metrics
    print(f"\nSmart Rate Limiter Metrics {async_label}:")
    for key, value in smart_limiter_metrics.items():
        if isinstance(value, float):
            print(f"{key}: {value:.2f}")
        else:
            print(f"{key}: {value}")

    print(f"\nPerformance Comparison {async_label}:")
    print("Success Rates:")
    print(f"  Baseline:   {(baseline_stats['successful']/total_prompts)*100:.1f}%")
    print(f"  Smart:      {(smart_stats['successful']/total_prompts)*100:.1f}%")
    print(f"  LangChain:  {(langchain_stats['successful']/total_prompts)*100:.1f}%")

    print("\nRejected Requests:")
    print(f"  Baseline:   {baseline_stats['rejected']}")
    print(f"  Smart:      {smart_stats['rejected']}")
    print(f"  LangChain:  {langchain_stats['rejected']}")

    print("\nTotal Successful Tokens:")
    print(f"  Baseline:   {baseline_stats['tokens']}")
    print(f"  Smart:      {smart_stats['tokens']}")
    print(f"  LangChain:  {langchain_stats['tokens']}")

    print("\nDuration (seconds):")
    print(f"  Baseline:   {baseline_stats['duration']:.2f}")
    print(f"  Smart:      {smart_stats['duration']:.2f}")
    print(f"  LangChain:  {langchain_stats['duration']:.2f}")

    print("\nSuccessful RPM:")
    print(
        f"  Baseline:   {baseline_stats['successful']/(baseline_stats['duration']/60):.1f}"
    )
    print(f"  Smart:      {smart_stats['successful']/(smart_stats['duration']/60):.1f}")
    print(
        f"  LangChain:  {langchain_stats['successful']/(langchain_stats['duration']/60):.1f}"
    )

    print("\nSuccessful TPM:")
    print(
        f"  Baseline:   {baseline_stats['tokens']/(baseline_stats['duration']/60):.1f}"
    )
    print(f"  Smart:      {smart_stats['tokens']/(smart_stats['duration']/60):.1f}")
    print(
        f"  LangChain:  {langchain_stats['tokens']/(langchain_stats['duration']/60):.1f}"
    )


def run_assertions(baseline_stats: Dict, smart_stats: Dict, langchain_stats: Dict):
    """Run assertions to verify rate limiter performance"""
    assert (
        smart_stats["successful"] >= baseline_stats["successful"]
    ), "Smart rate limiter should improve success rate"
    assert (
        langchain_stats["successful"] >= baseline_stats["successful"]
    ), "LangChain rate limiter should improve success rate"
    assert (
        smart_stats["rejected"] <= baseline_stats["rejected"]
    ), "Smart rate limiter should reduce rejections"
    assert (
        langchain_stats["rejected"] <= baseline_stats["rejected"]
    ), "LangChain rate limiter should reduce rejections"


def collect_test_stats(results: List, duration: float, api: MockLLMAPI) -> Dict:
    """Collect statistics from a test run"""
    successful_results = [r for r in results if isinstance(r, MockLLMResponse)]
    return {
        "successful": len(successful_results),
        "tokens": sum(r.tokens_used for r in successful_results),
        "rejected": api.rejected_requests,
        "duration": duration,
    }


def test_smart_rate_limiter_with_mock_api():
    """Test SmartRateLimiter with a mock LLM API under heavy load

    MockLLM API limits:
    - 60 requests per minute
    - 4000 tokens per minute

    Test processing 150 prompts of varying lengths with 10 workers
    This is to simulate a load higher than the LLM API can handle
    - We are making 10 requests per second to the API that can only handle 1 request per second.
    - If a request to the API fails (i.e we exceed rate limits), we count it as a rejected request.
    - We then compare the success rate, rejected requests, and processing time of the different rate limiters.

    We compare 3 scenarios:
    - Baseline: No rate limiter
    - Smart rate limiter (using requests and tokens)
    - LangChain rate limiter (In Memory Rate Limiter using only requests to control)

    Ideally, the smart rate limiter should improve the success rate and keep the RPM and TPM close to the API limits.
    """
    # Setup
    api = MockLLMAPI(rpm_limit=60, tpm_limit=4000)  # 1 request/sec, ~66 tokens/sec
    max_workers = 10
    enc = tiktoken.encoding_for_model("gpt-4")
    tokenizer = lambda x: len(enc.encode(x))

    smart_limiter = SmartRateLimiter(
        requests_per_minute=60,
        tokens_per_minute=4000,
        max_request_burst=10,
        tokenizer=tokenizer,
        check_every_n_seconds=0.1,
    )

    langchain_limiter = InMemoryRateLimiter(
        requests_per_second=1,  # <-- Can only make a request once every 10 seconds!!
        check_every_n_seconds=0.1,  # Wake up every 100 ms to check whether allowed to make a request,
        max_bucket_size=10,  # Controls the maximum burst size.
    )

    # Test prompts of varying lengths - increased volume and variety
    prompts = [
        "Short prompt",  # ~2 tokens
        "This is a medium length prompt that uses more tokens than the short one",  # ~12 tokens
        "This is a much longer prompt that will use even more tokens and really test our rate limiting capabilities with many more words and complex ideas",  # ~27 tokens
    ] * 5  # 15 prompts total, should exceed rate limits

    def process_with_smart_limiter(prompt: str):
        smart_limiter.acquire(prompt)
        try:
            return api.process_prompt(prompt)
        except Exception:
            return None

    def process_with_langchain_limiter(prompt: str):
        langchain_limiter.acquire()
        try:
            return api.process_prompt(prompt)
        except Exception:
            return None

    def process_without_limiter(prompt: str):
        try:
            return api.process_prompt(prompt)
        except Exception:
            return None

    # Test without rate limiter (baseline)
    api.total_requests = 0
    api.rejected_requests = 0
    start_time = time.time()

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        results = list(executor.map(process_without_limiter, prompts))

    baseline_stats = collect_test_stats(results, time.time() - start_time, api)
    print_test_metrics(
        "baseline",
        baseline_stats["duration"],
        baseline_stats["successful"],
        len(prompts),
        api,
        baseline_stats["tokens"],
    )

    # Test with smart rate limiter
    api = MockLLMAPI(rpm_limit=60, tpm_limit=4000)
    start_time = time.time()

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        results = list(executor.map(process_with_smart_limiter, prompts))

    smart_stats = collect_test_stats(results, time.time() - start_time, api)
    print_test_metrics(
        "smart rate limiter",
        smart_stats["duration"],
        smart_stats["successful"],
        len(prompts),
        api,
        smart_stats["tokens"],
    )

    # Test with LangChain rate limiter
    api = MockLLMAPI(rpm_limit=60, tpm_limit=4000)
    start_time = time.time()

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        results = list(executor.map(process_with_langchain_limiter, prompts))

    langchain_stats = collect_test_stats(results, time.time() - start_time, api)
    print_test_metrics(
        "LangChain rate limiter",
        langchain_stats["duration"],
        langchain_stats["successful"],
        len(prompts),
        api,
        langchain_stats["tokens"],
    )

    # Print comparison and run assertions
    print_comparison_metrics(
        baseline_stats,
        smart_stats,
        langchain_stats,
        len(prompts),
        smart_limiter.get_metrics(),
    )
    run_assertions(baseline_stats, smart_stats, langchain_stats)


@pytest.mark.asyncio
async def test_smart_rate_limiter_with_mock_api_async():
    """Test async SmartRateLimiter with a mock LLM API under heavy load

    This test is identical to test_smart_rate_limiter_with_mock_api but uses async methods.
    It tests the aacquire() methods of both rate limiters.
    """
    # Setup
    api = MockLLMAPI(rpm_limit=60, tpm_limit=4000)
    max_concurrent = 10
    enc = tiktoken.encoding_for_model("gpt-4")
    tokenizer = lambda x: len(enc.encode(x))

    smart_limiter = SmartRateLimiter(
        requests_per_minute=60,
        tokens_per_minute=4000,
        max_request_burst=10,
        tokenizer=tokenizer,
        check_every_n_seconds=0.1,
    )

    langchain_limiter = InMemoryRateLimiter(
        requests_per_second=1,
        check_every_n_seconds=0.1,
        max_bucket_size=10,
    )

    # Test prompts
    prompts = [
        "Short prompt",
        "This is a medium length prompt that uses more tokens than the short one",
        "This is a much longer prompt that will use even more tokens and really test our rate limiting capabilities with many more words and complex ideas",
    ] * 5

    async def process_with_smart_limiter(prompt: str):
        await smart_limiter.aacquire(prompt)
        try:
            return await api.aprocess_prompt(prompt)
        except Exception:
            return None

    async def process_with_langchain_limiter(prompt: str):
        await langchain_limiter.aacquire()
        try:
            return await api.aprocess_prompt(prompt)
        except Exception:
            return None

    async def process_without_limiter(prompt: str):
        try:
            return await api.aprocess_prompt(prompt)
        except Exception:
            return None

    async def run_concurrent(func, prompts):
        tasks = [func(prompt) for prompt in prompts]
        return await asyncio.gather(*tasks, return_exceptions=True)

    # Test without rate limiter (baseline)
    api.total_requests = 0
    api.rejected_requests = 0
    start_time = time.time()

    results = await run_concurrent(process_without_limiter, prompts)

    baseline_stats = collect_test_stats(results, time.time() - start_time, api)
    print_test_metrics(
        "baseline (async)",
        baseline_stats["duration"],
        baseline_stats["successful"],
        len(prompts),
        api,
        baseline_stats["tokens"],
    )

    # Test with smart rate limiter
    api = MockLLMAPI(rpm_limit=60, tpm_limit=4000)
    start_time = time.time()

    results = await run_concurrent(process_with_smart_limiter, prompts)

    smart_stats = collect_test_stats(results, time.time() - start_time, api)
    print_test_metrics(
        "smart rate limiter (async)",
        smart_stats["duration"],
        smart_stats["successful"],
        len(prompts),
        api,
        smart_stats["tokens"],
    )

    # Test with LangChain rate limiter
    api = MockLLMAPI(rpm_limit=60, tpm_limit=4000)
    start_time = time.time()

    results = await run_concurrent(process_with_langchain_limiter, prompts)

    langchain_stats = collect_test_stats(results, time.time() - start_time, api)
    print_test_metrics(
        "LangChain rate limiter (async)",
        langchain_stats["duration"],
        langchain_stats["successful"],
        len(prompts),
        api,
        langchain_stats["tokens"],
    )

    # Print comparison and run assertions
    print_comparison_metrics(
        baseline_stats,
        smart_stats,
        langchain_stats,
        len(prompts),
        smart_limiter.get_metrics(),
        is_async=True,
    )
    run_assertions(baseline_stats, smart_stats, langchain_stats)



from pydantic import BaseModel
from typing import TypeVar
import logging
from logging import getLogger, StreamHandler
import os
import sys
import hashlib
from uuid import UUID


def get_logger(name: str = None):
    logger = getLogger(name or __name__)
    logger.setLevel(os.getenv("LOG_LEVEL", "DEBUG"))

    # allow stdout to be logged
    logger.addHandler(logging.StreamHandler(sys.stdout))

    # log format
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    logger.handlers[0].setFormatter(formatter)
    return logger


logger = get_logger(__name__)

T = TypeVar("T")  # Generic type for input
R = TypeVar("R")  # Generic type for result
BaseModelType = TypeVar("BaseModelType", bound=BaseModel)


def decider(uuid: UUID, sample_rate: float, hash_salt: str = "") -> bool:
    """
    Deterministically decide whether to sample a UUID based on sample rate.

    Args:
        uuid: UUID to check
        sample_rate: Float between 0 and 1 representing sampling probability
        hash_salt: Optional salt string to mix into hash

    Returns:
        bool: True if UUID should be sampled based on rate, False otherwise
    """
    if sample_rate <= 0:
        return False
    if sample_rate >= 1:
        return True

    # Create hash of UUID + salt
    hash_input = f"{uuid}{hash_salt}".encode("utf-8")
    hash_value = hashlib.md5(hash_input).hexdigest()

    # Convert first 4 bytes of hash to int and normalize to 0-1
    hash_int = int(hash_value[:8], 16)
    normalized = hash_int / 0xFFFFFFFF

    return normalized < sample_rate



import pytest
from unittest.mock import patch, MagicMock
from langchain_llm_utils.batch_process import (
    batch_process_with_progress,
    BatchProcessConfig,
    WorkloadType,
)
from langchain_llm_utils.llm import LLM


@pytest.fixture
def mock_llm_provider():
    """Fixture to mock the LLM provider."""
    with patch(
        "langchain_llm_utils.llm.LLMFactory.create_provider"
    ) as mock_create_provider:
        mock_provider = MagicMock()
        mock_create_provider.return_value = mock_provider
        yield mock_provider


def test_batch_process_with_progress_success(mock_llm_provider):
    """Test successful batch processing."""
    # Mock the _llm.invoke method to return processed items
    mock_llm_provider.invoke.side_effect = lambda x: f"Processed {x}"
    items = ["item1", "item2", "item3"]
    config = BatchProcessConfig(workload_type=WorkloadType.CPU_BOUND)

    results = batch_process_with_progress(items, mock_llm_provider.invoke, config)

    assert results == ["Processed item1", "Processed item2", "Processed item3"]
    assert mock_llm_provider.invoke.call_count == len(items)


def test_batch_process_with_progress_error_handling(mock_llm_provider):
    """Test error handling in batch processing."""
    # Mock the _llm.invoke method to return None, empty string, or return a processed item
    mock_llm_provider.invoke.side_effect = [
        None,  # Simulate a None response for item1
        "",  # Simulate an empty string for item2
        "Processed item3",
    ]
    items = ["item1", "item2", "item3"]
    config = BatchProcessConfig(workload_type=WorkloadType.CPU_BOUND)

    results = batch_process_with_progress(items, mock_llm_provider.invoke, config)

    # Check that the results contain None for the failed item
    assert results == [None, "", "Processed item3"]
    assert mock_llm_provider.invoke.call_count == len(items)


def test_batch_process_with_progress_io_bound(mock_llm_provider):
    """Test batch processing with IO_BOUND workload type."""
    # Mock the _llm.invoke method to return processed items
    mock_llm_provider.invoke.side_effect = lambda x: f"Processed {x}"
    items = ["item1", "item2", "item3"]
    config = BatchProcessConfig(workload_type=WorkloadType.IO_BOUND)

    results = batch_process_with_progress(items, mock_llm_provider.invoke, config)

    assert results == ["Processed item1", "Processed item2", "Processed item3"]
    assert mock_llm_provider.invoke.call_count == len(items)


def test_batch_process_with_progress_mixed_workload(mock_llm_provider):
    """Test batch processing with MIXED workload type."""
    # Mock the _llm.invoke method to return processed items
    mock_llm_provider.invoke.side_effect = lambda x: f"Processed {x}"
    items = ["item1", "item2", "item3"]
    config = BatchProcessConfig(workload_type=WorkloadType.MIXED, cpu_allocation=0.5)

    results = batch_process_with_progress(items, mock_llm_provider.invoke, config)

    assert results == ["Processed item1", "Processed item2", "Processed item3"]
    assert mock_llm_provider.invoke.call_count == len(items)
